"""
This module launches and controls the running of MPI applications.

In order to create an MPI executor, the calling script should contain ::

    exctr = MPIExecutor()

See the executor API below for optional arguments.
"""

import os
import logging
import time

import libensemble.utils.launcher as launcher
from libensemble.resources.mpi_resources import get_MPI_variant
from libensemble.executors.executor import Executor, Task, ExecutorException
from libensemble.executors.mpi_runner import MPIRunner

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


class MPIExecutor(Executor):
    """The MPI executor can create, poll and kill runnable MPI tasks
    """

    def __init__(self, custom_info={}):
        """Instantiate a new MPIExecutor instance.

        A new MPIExecutor is created with an application
        registry and configuration attributes.

        This is typically created in the user calling script. The
        MPIExecutor will use system resource information supplied by
        the libEsnemble resource manager when submitting tasks.

        Parameters
        ----------

        custom_info: dict, optional
            Provide custom overrides to selected variables that are usually
            auto-detected. See below.


        The MPIExecutor automatically detects MPI runners and launch
        mechanisms. However it is possible to override the detected
        information using the ``custom_info`` argument. This takes
        a dictionary of values.

        The allowable fields are::

            'mpi_runner' [string]:
                Select runner: 'mpich', 'openmpi', 'aprun', 'srun', 'jsrun', 'custom'
                All except 'custom' relate to runner classes in libEnsemble.
                Custom allows user to define their own run-lines but without parsing
                arguments or making use of auto-resources.
            'runner_name' [string]:
                Runner name: Replaces run command if present. All runners have a default
                except for 'custom'.
            'subgroup_launch' [Boolean]:
                Whether MPI runs should be initiatied in a new process group. This needs
                to be correct for kills to work correctly. Use the standalone test at
                libensemble/tests/standalone_tests/kill_test to determine correct value
                for a system.

        For example::

            customizer = {'mpi_runner': 'mpich',
                          'runner_name': 'wrapper -x mpich'}

            from libensemble.executors.mpi_executor import MPIExecutor
            exctr = MPIExecutor(custom_info=customizer)

        """

        Executor.__init__(self)

        # MPI launch settings
        self.max_launch_attempts = 5
        self.fail_time = 2
        self.retry_delay_incr = 5  # Incremented wait after each launch attempt

        # Apply custom options
        mpi_runner_type = custom_info.get('mpi_runner', None)
        runner_name = custom_info.get('runner_name', None)
        subgroup_launch = custom_info.get('subgroup_launch', None)

        if not mpi_runner_type:
            mpi_runner_type = get_MPI_variant()
        self.mpi_runner = MPIRunner.get_runner(mpi_runner_type, runner_name)
        if subgroup_launch is not None:
            self.mpi_runner.subgroup_launch = subgroup_launch
        self.resources = None

    def set_resources(self, resources):
        self.resources = resources

    def _launch_with_retries(self, task, runline, subgroup_launch, wait_on_start):
        """ Launch task with retry mechanism"""
        retry_count = 0
        while retry_count < self.max_launch_attempts:
            retry = False
            try:
                retry_string = " (Retry {})".format(retry_count) if retry_count > 0 else ""
                logger.info("Launching task {}{}: {}".
                            format(task.name, retry_string, " ".join(runline)))
                task.run_attempts += 1
                with open(task.stdout, 'w') as out, open(task.stderr, 'w') as err:
                    task.process = launcher.launch(runline, cwd='./',
                                                   stdout=out,
                                                   stderr=err,
                                                   start_new_session=subgroup_launch)
            except Exception as e:
                logger.warning('task {} submit command failed on '
                               'try {} with error {}'
                               .format(task.name, retry_count, e))
                retry = True
                retry_count += 1
            else:
                if (wait_on_start):
                    self._wait_on_start(task, self.fail_time)

                if task.state == 'FAILED':
                    logger.warning('task {} failed within fail_time on '
                                   'try {} with err code {}'
                                   .format(task.name, retry_count, task.errcode))
                    retry = True
                    retry_count += 1

            if retry and retry_count < self.max_launch_attempts:
                logger.debug('Retry number {} for task {}'.format(retry_count, task.name))
                time.sleep(retry_count*self.retry_delay_incr)
                task.reset()  # Some cases may require user cleanup
            else:
                break

    def submit(self, calc_type=None, app_name=None, num_procs=None,
               num_nodes=None, procs_per_node=None, machinefile=None,
               app_args=None, stdout=None, stderr=None, stage_inout=None,
               hyperthreads=False, dry_run=False, wait_on_start=False,
               extra_args=None):
        """Creates a new task, and either executes or schedules execution.

        The created task object is returned.

        Parameters
        ----------

        calc_type: String, optional
            The calculation type: 'sim' or 'gen'
            Only used if app_name is not supplied. Uses default sim or gen application.

        app_name: String, optional
            The application name. Must be supplied if calc_type is not.

        num_procs: int, optional
            The total number of MPI tasks on which to submit the task

        num_nodes: int, optional
            The number of nodes on which to submit the task

        procs_per_node: int, optional
            The processes per node for this task

        machinefile: string, optional
            Name of a machinefile for this task to use

        app_args: string, optional
            A string of the application arguments to be added to task
            submit command line

        stdout: string, optional
            A standard output filename

        stderr: string, optional
            A standard error filename

        stage_inout: string, optional
            A directory to copy files from; default will take from
            current directory

        hyperthreads: boolean, optional
            Whether to submit MPI tasks to hyperthreads

        dry_run: boolean, optional
            Whether this is a dry_run - no task will be launched; instead
            runline is printed to logger (at INFO level)

        wait_on_start: boolean, optional
            Whether to wait for task to be polled as RUNNING (or other
            active/end state) before continuing

        extra_args: String, optional
            Additional command line arguments to supply to MPI runner. If
            arguments are recognised as MPI resource configuration
            (num_procs, num_nodes, procs_per_node) they will be used in
            resources determination unless also supplied in the direct
            options.

        Returns
        -------

        task: obj: Task
            The launched task object


        Note that if some combination of num_procs, num_nodes, and
        procs_per_node is provided, these will be honored if
        possible. If resource detection is on and these are omitted,
        then the available resources will be divided among workers.
        """

        if app_name is not None:
            app = self.get_app(app_name)
        elif calc_type is not None:
            app = self.default_app(calc_type)
        else:
            raise ExecutorException("Either app_name or calc_type must be set")

        default_workdir = os.getcwd()
        task = Task(app, app_args, default_workdir, stdout, stderr, self.workerID)

        if stage_inout is not None:
            logger.warning("stage_inout option ignored in this "
                           "executor - runs in-place")

        mpi_specs = self.mpi_runner.get_mpi_specs(task, num_procs, num_nodes,
                                                  procs_per_node, machinefile,
                                                  hyperthreads, extra_args,
                                                  self.resources,
                                                  self.workerID)

        mpi_command = self.mpi_runner.mpi_command
        sglaunch = self.mpi_runner.subgroup_launch
        runline = launcher.form_command(mpi_command, mpi_specs)

        runline.extend(task.app.full_path.split())
        if task.app_args is not None:
            runline.extend(task.app_args.split())

        task.runline = ' '.join(runline)  # Allow to be queried
        if dry_run:
            task.dry_run = True
            logger.info('Test (No submit) Runline: {}'.format(' '.join(runline)))
            task._set_complete(dry_run=True)
        else:
            # Launch Task
            self._launch_with_retries(task, runline, sglaunch, wait_on_start)

            if not task.timer.timing:
                task.timer.start()
                task.submit_time = task.timer.tstart  # Time not date - may not need if using timer.

        self.list_of_tasks.append(task)
        return task

    def set_worker_info(self, comm, workerid=None):
        """Sets info for this executor"""
        super().set_worker_info(comm, workerid)



#!/usr/bin/env python

"""
Unit test of libensemble log functions.
"""
import os
import logging
from libensemble import logger
from libensemble.comms.logs import LogConfig


def test_set_log_level():
    # Default
    level = logger.get_level()
    assert level == 20, "Log level should be 20. Found: " + str(level)

    logger.set_level('DEBUG')
    level = logger.get_level()
    assert level == 10, "Log level should be 10. Found: " + str(level)

    logger.set_level('WARNING')
    level = logger.get_level()
    assert level == 30, "Log level should be 30. Found: " + str(level)

    logger.set_level('MANAGER_WARNING')
    level = logger.get_level()
    assert level == 35, "Log level should be 35. Found: " + str(level)

    logger.set_level('ERROR')
    level = logger.get_level()
    assert level == 40, "Log level should be 40. Found: " + str(level)

    logger.set_level('INFO')
    level = logger.get_level()
    assert level == 20, "Log level should be 20. Found: " + str(level)


# def test_change_log_level():
#     from libensemble.comms.logs import manager_logging_config
#     manager_logging_config()
#     level_from_config = logger.get_level()
#     assert level_from_config == 20, "Log level from config should be 20. Found: " + str(level)
#     level_from_logger = logging.getLogger('libensemble').getEffectiveLevel()
#     assert level_from_logger == 20, "Log level from logger should be 20. Found: " + str(level)

#     # Now test logger level after change
#     logger.set_level('DEBUG')
#     level_from_logger = logging.getLogger('libensemble').getEffectiveLevel()
#     assert level_from_logger == 10, "Log level from logger should be 10. Found: " + str(level)

#     # Now test inherited logger
#     level_from_child_logger = logging.getLogger('libensemble.libE').getEffectiveLevel()
#     assert level_from_child_logger == 10, "Log level from child logger should be 10. Found: " + str(level)


def test_set_filename():
    from libensemble.comms.logs import manager_logging_config
    alt_name = "alt_name.log"

    logs = LogConfig.config
    assert logs.filename == "ensemble.log", "Log filename expected ensemble.log. Found: " + logs.filename

    logger.set_filename(alt_name)
    assert logs.filename == alt_name, "Log filename expected " + str(alt_name) + ". Found: " + logs.filename

    manager_logging_config()
    logger.set_filename('toolate.log')
    assert logs.filename == alt_name, "Log filename expected " + str(alt_name) + ". Found: " + logs.filename

    assert os.path.isfile(alt_name), "Expected creation of file" + str(alt_name)
    with open(alt_name, 'r') as f:
        line = f.readline()
        assert "Cannot set filename after loggers initialized" in line
    os.remove(alt_name)

    logs = LogConfig.config
    logs.logger_set = True
    logs.set_level('DEBUG')


def test_set_stderr_level():

    stderr_level = logger.get_stderr_level()
    assert stderr_level == 35, "Default stderr copying level is 35, found " + \
        str(stderr_level)

    logger.set_stderr_level('DEBUG')
    stderr_level = logger.get_stderr_level()
    assert stderr_level == 10, "Log level should be 10. Found: " + str(stderr_level)

    logger.set_stderr_level('INFO')
    stderr_level = logger.get_stderr_level()
    assert stderr_level == 20, "Log level should be 20. Found: " + str(stderr_level)

    logger.set_stderr_level('WARNING')
    stderr_level = logger.get_stderr_level()
    assert stderr_level == 30, "Log level should be 30. Found: " + str(stderr_level)

    logger.set_stderr_level('MANAGER_WARNING')
    stderr_level = logger.get_stderr_level()
    assert stderr_level == 35, "Log level should be 35. Found: " + str(stderr_level)

    logger.set_stderr_level('ERROR')
    stderr_level = logger.get_stderr_level()
    assert stderr_level == 40, "Log level should be 40. Found: " + str(stderr_level)

    logger.set_level('ERROR')
    logger_test = logging.getLogger('libensemble')
    logger_test.manager_warning('This test message should not log')


# Need setup/teardown here to kill loggers if running file without pytest
# Issue: cannot destroy loggers and they are set up in other unit tests.
# Partial solution: either rename the file so it is the first unit test, or
#   move this unit test to its own directory.
if __name__ == "__main__":
    test_set_log_level()
    # test_change_log_level()
    test_set_filename()
    test_set_stderr_level()



import sys
import pytest
import numpy as np
from libensemble.resources.scheduler import (ResourceScheduler,
                                             InsufficientFreeResources,
                                             InsufficientResourcesError)


class MyResources:
    """Simulate resources"""

    rset_dtype = [('assigned', int),  # Holds worker ID assigned to or zero
                  ('group', int)]     # Group ID this resource set belongs to

    # Basic layout
    def __init__(self, num_rsets, num_groups):
        self.total_num_rsets = num_rsets
        self.num_groups = num_groups
        self.rsets_per_node = self.total_num_rsets//num_groups
        self.even_groups = True
        self.rsets = np.zeros(self.total_num_rsets, dtype=MyResources.rset_dtype)
        self.rsets['assigned'] = 0
        for i in range(self.total_num_rsets):
            self.rsets['group'][i] = i // self.rsets_per_node
        self.rsets_free = self.total_num_rsets
        print(self.rsets)

    def free_rsets(self, worker=None):
        """Free up assigned resource sets"""
        if worker is None:
            self.rsets['assigned'] = 0
            self.rsets_free = self.total_num_rsets
        else:
            for rset, wid in enumerate(self.rsets['assigned']):
                if wid == worker:
                    self.rsets['assigned'][rset] = 0
                    self.rsets_free += 1

    def assign_rsets(self, rset_team, worker_id):
        """Mark the resource sets given by rset_team as assigned to worker_id"""
        if rset_team:
            self.rsets['assigned'][rset_team] = worker_id
            self.rsets_free -= len(rset_team)  # quick count

    # Special function for testing from a given starting point
    def fixed_assignment(self, assignment):
        """Set the given assignment along with other coupled information"""
        self.rsets['assigned'] = assignment
        self.rsets_free = np.count_nonzero(self.rsets['assigned'] == 0)


def test_too_many_rsets():
    """Tests request of more resource sets than exist"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    sched = ResourceScheduler(user_resources=resources)

    with pytest.raises(InsufficientResourcesError):
        rset_team = sched.assign_resources(rsets_req=10)  # noqa F841
        pytest.fail('Expected InsufficientResourcesError')


def test_cannot_split_quick_return():
    """Tests the quick return when splitting finds no free even gaps"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(6, 3)
    resources.fixed_assignment(([1, 0, 0, 0, 3, 3]))
    sched = ResourceScheduler(user_resources=resources)

    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=3)
        print(rset_team)
        pytest.fail('Expected InsufficientFreeResources')


def test_schdule_find_gaps_1node():
    """Tests assignment of rsets on one node.

    This test also checks the list is correctly assigned to workers
    and the freeing of assigned resources.
    """
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 1)
    sched = ResourceScheduler(user_resources=resources)

    rset_team = sched.assign_resources(rsets_req=2)
    assert rset_team == [0, 1]

    rset_team = sched.assign_resources(rsets_req=3)
    assert rset_team == [2, 3, 4]

    # Check not enough slots
    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=4)
        pytest.fail('Expected InsufficientFreeResources')

    rset_team = sched.assign_resources(rsets_req=2)
    assert rset_team == [5, 6]

    # Simulate resources freed up on return from worker
    resources.fixed_assignment(([3, 3, 0, 0, 0, 4, 4, 0]))

    # Create new scheduler to simulate new alloc call
    del sched
    sched = ResourceScheduler(user_resources=resources)

    rset_team = sched.assign_resources(rsets_req=4)
    assert rset_team == [2, 3, 4, 7]
    del resources


def test_schdule_find_gaps_2nodes():
    """Tests finding gaps on two nodes with equal resource sets"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    sched = ResourceScheduler(user_resources=resources)
    inputs = [2, 3, 1, 2]
    exp_out = [[0, 1], [4, 5, 6], [7], [2, 3]]
    for i in range(4):
        rset_team = sched.assign_resources(rsets_req=inputs[i])
        assert rset_team == exp_out[i], \
            'Expected {}, Received rset_team {}'.format(exp_out[i], rset_team)

    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=1)
        pytest.fail('Expected InsufficientFreeResources')

    del resources


def test_across_nodes_even_split():
    """Tests assignment over two nodes"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    sched = ResourceScheduler(user_resources=resources)
    rset_team = sched.assign_resources(rsets_req=6)
    # Expecting even split
    assert rset_team == [0, 1, 2, 4, 5, 6], \
        'Even split test did not get expected result {}'.format(rset_team)

    # This time it must use 3 nodes for even split (though 2 would cover uneven).
    resources = MyResources(15, 3)
    sched = ResourceScheduler(user_resources=resources)
    rset_team = sched.assign_resources(rsets_req=9)
    # Expecting even split
    assert rset_team == [0, 1, 2, 5, 6, 7, 10, 11, 12], \
        'Even split test did not get expected result {}'.format(rset_team)
    del resources


def test_across_nodes_roundup_option():
    """Tests assignment over two nodes"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    sched = ResourceScheduler(user_resources=resources)
    rset_team = sched.assign_resources(rsets_req=5)
    # Expecting even split
    assert rset_team == [0, 1, 2, 4, 5, 6], \
        'Even split test did not get expected result {}'.format(rset_team)
    del resources


def test_try1node_findon_2nodes():
    """Tests finding gaps on two nodes as cannot fit on one due to others assigned"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    resources.fixed_assignment(([1, 1, 0, 0, 0, 2, 2, 0]))
    sched = ResourceScheduler(user_resources=resources)
    rset_team = sched.assign_resources(rsets_req=4)
    assert rset_team == [2, 3, 4, 7], 'rsets found {}'.format(rset_team)
    del resources


def test_try1node_findon_3nodes():
    """Tests finding gaps on two nodes as cannot fit on one due to others assigned"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(12, 3)
    resources.fixed_assignment(([1, 1, 0, 0, 0, 2, 2, 0, 3, 0, 3, 3]))
    sched = ResourceScheduler(user_resources=resources)

    rset_team = sched.assign_resources(rsets_req=3)
    assert rset_team == [2, 4, 9], 'rsets found {}'.format(rset_team)

    # Simulate a new call to allocation function
    sched_options = {'split2fit': False}
    del sched; sched = ResourceScheduler(user_resources=resources, sched_opts=sched_options)  # noqa E702

    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=3)
        pytest.fail('Expected InsufficientFreeResources')

    # Now free up resources on 1st group and call alloc again (new sched).
    resources.free_rsets(worker=1)
    del sched; sched = ResourceScheduler(user_resources=resources, sched_opts=sched_options)  # noqa E702
    rset_team = sched.assign_resources(rsets_req=3)
    assert rset_team == [0, 1, 2], 'rsets found {}'.format(rset_team)
    del resources


def test_try2nodes_findon_3nodes():
    """Tests finding gaps on two nodes as cannot fit on one due to others assigned"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(18, 3)
    resources.fixed_assignment(([0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3]))
    sched = ResourceScheduler(user_resources=resources)

    # Can't find 2 groups of 6 so find 3 groups of 4.
    rset_team = sched.assign_resources(rsets_req=12)
    assert rset_team == [0, 2, 3, 4, 6, 7, 8, 9, 12, 13, 14, 15], 'rsets found {}'.format(rset_team)

    # Simulate a new call to allocation function
    sched_options = {'split2fit': False}
    del sched; sched = ResourceScheduler(user_resources=resources, sched_opts=sched_options)  # noqa E702
    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=12)
        pytest.fail('Expected InsufficientFreeResources')

    # Now free up resources on 3rd group and call alloc again (new sched).
    resources.free_rsets(worker=3)
    del sched; sched = ResourceScheduler(user_resources=resources, sched_opts=sched_options)  # noqa E702
    rset_team = sched.assign_resources(rsets_req=12)
    assert rset_team == [6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17], 'rsets found {}'.format(rset_team)
    del resources


def test_split2fit_even_required_fails():
    """Tests tries one node then two, and both fail"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    resources.fixed_assignment(([1, 1, 1, 0, 2, 2, 0, 0]))
    sched = ResourceScheduler(user_resources=resources)
    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=4)  # noqa F841
        pytest.fail('Expected InsufficientFreeResources')


def test_split2fit_even_required_various():
    """Tests trying to fit to an non-even partition, and setting of local rsets_free"""
    print("\nTest: {}\n".format(sys._getframe().f_code.co_name))
    resources = MyResources(8, 2)
    resources.fixed_assignment(([1, 1, 1, 0, 0, 0, 0, 0]))
    sched = ResourceScheduler(user_resources=resources)
    assert sched.rsets_free == 5

    rset_team = sched.assign_resources(rsets_req=2)
    assert rset_team == [4, 5], 'rsets found {}'.format(rset_team)
    assert sched.rsets_free == 3

    # In same alloc - now try getting 4 rsets
    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=4)
        pytest.fail('Expected InsufficientFreeResources')
    assert sched.rsets_free == 3

    with pytest.raises(InsufficientFreeResources):
        rset_team = sched.assign_resources(rsets_req=3)
        pytest.fail('Expected InsufficientFreeResources')
    assert sched.rsets_free == 3

    rset_team = sched.assign_resources(rsets_req=2)
    assert rset_team == [6, 7], 'rsets found {}'.format(rset_team)
    assert sched.rsets_free == 1


if __name__ == "__main__":
    test_too_many_rsets()
    test_cannot_split_quick_return()
    test_schdule_find_gaps_1node()
    test_schdule_find_gaps_2nodes()
    test_across_nodes_even_split()
    test_across_nodes_roundup_option()
    test_try1node_findon_2nodes()
    test_try1node_findon_3nodes()
    test_try2nodes_findon_3nodes()
    test_split2fit_even_required_fails()
    test_split2fit_even_required_various()



"""
Tests the ability of libEnsemble to
 - give back all of the history to a persistent gen at shutdown

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_uniform_sampling_adv.py
   python3 test_persistent_uniform_sampling_adv.py --nworkers 3 --comms local
   python3 test_persistent_uniform_sampling_adv.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['use_persis_return_gen'] = True

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('grad', float, n)],
}

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'x', 'grad', 'sim_id'],
    'out': [('x', float, (n,))],
    'user': {
        'initial_batch_size': 100,
        'replace_final_fields': True,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

sim_max = 40
exit_criteria = {'sim_max': 40}

alloc_specs = {'alloc_f': alloc_f, 'out': []}

libE_specs['final_fields'] = ['x', 'f', 'sim_id']
# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert len(np.unique(H['gen_time'])) == 1, "Everything should have been generated in one batch"
    assert np.all(H['x'][0:sim_max] == -1.23), "The persistent gen should have set these at shutdown"

    save_libE_output(H, persis_info, __file__, nworkers)



"""
Runs libEnsemble on a gen_f that is missing necessary information; this tests
libE worker exception raising

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_worker_exceptions.py
   python3 test_worker_exceptions.py --nworkers 3 --comms local
   python3 test_worker_exceptions.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

from libensemble.libE import libE
from libensemble.tests.regression_tests.support import nan_func as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.manager import LoggedException
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()
n = 2

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': [],
    'out': [('x', float, 2)],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'initial_sample': 100,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

libE_specs['abort_on_exception'] = False

# Tell libEnsemble when to stop
exit_criteria = {'elapsed_wallclock_time': 10}

# Perform the run
return_flag = 1
try:
    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)
except LoggedException as e:
    print("Caught deliberate exception: {}".format(e))
    return_flag = 0

if is_manager:
    assert return_flag == 0



"""
Below are the fields used within libEnsemble
"""

libE_fields = [('sim_id', int),             # Unique id of entry in H that was generated
               ('gen_worker', int),         # Worker that (first) generated the entry
               ('gen_time', float),         # Time (since epoch) entry (first) was entered into H from a gen
               ('last_gen_time', float),    # Time (since epoch) entry was last requested by a gen
               ('given', bool),             # True if entry has been given for sim eval
               ('given_time', float),       # Time (since epoch) that the entry was (first) given to be evaluated
               ('last_given_time', float),  # Time (since epoch) that the entry was last given to be evaluated
               ('last_given_back_time', float),  # Time (since epoch) that the entry was last given back to the gen
               ('returned', bool),          # True if entry has been returned from sim eval
               ('returned_time', float),    # Time entry was (last) returned from sim eval
               ('sim_worker', int),         # Worker that did (or is doing) the sim eval
               ('cancel_requested', bool),  # True if cancellation of this entry is requested
               ('kill_sent', bool),         # True if a kill signal has been sent to worker
               ('given_back', bool)         # True if entry has been given back to the gen after evaluation.
               ]
# end_libE_fields_rst_tag

protected_libE_fields = ['gen_worker',
                         'gen_time',
                         'given',
                         'given_time',
                         'returned',
                         'returned_time',
                         'given_back',
                         'sim_worker']

allowed_sim_spec_keys = ['sim_f',  #
                         'in',     #
                         'persis_in',
                         'out',    #
                         'user']   #

allowed_gen_spec_keys = ['gen_f',  #
                         'in',     #
                         'persis_in',
                         'out',    #
                         'user']   #

allowed_alloc_spec_keys = ['alloc_f',  #
                           'out',      #
                           'user']     #

libE_spec_calc_dir_misc = ['ensemble_copy_back',
                           'ensemble_dir_path',
                           'use_worker_dirs']

libE_spec_sim_dir_keys = ['sim_dirs_make',
                          'sim_dir_copy_files',
                          'sim_dir_symlink_files',
                          'sim_input_dir']

libE_spec_gen_dir_keys = ['gen_dirs_make',
                          'gen_dir_copy_files',
                          'gen_dir_symlink_files',
                          'gen_input_dir']

libE_spec_calc_dir_combined = libE_spec_calc_dir_misc + \
    libE_spec_sim_dir_keys + \
    libE_spec_gen_dir_keys

allowed_libE_spec_keys = ['abort_on_exception',             #
                          'enforce_worker_core_bounds',     #
                          'authkey',                        #
                          'disable_resource_manager',       #
                          'dedicated_mode',                   #
                          'comms',                          #
                          'resource_info',                  #
                          'disable_log_files',              #
                          'final_fields',                   #
                          'ip',                             #
                          'kill_canceled_sims',             #
                          'mpi_comm',                       #
                          'num_resource_sets',              #
                          'nworkers',                       #
                          'port',                           #
                          'profile',                        #
                          'safe_mode',                      #
                          'save_every_k_gens',              #
                          'save_every_k_sims',              #
                          'save_H_and_persis_on_abort',     #
                          'use_persis_return_gen',          #
                          'use_persis_return_sim',          #
                          'workerID',                       #
                          'worker_timeout',                 #
                          'zero_resource_workers',          #
                          'worker_cmd'] + libE_spec_calc_dir_combined



__all__ = ['run_keras_cvae_ml_genf']

import os
import glob
import yaml
import json
import numpy as np
from libensemble.executors.executor import Executor
from libensemble.message_numbers import (STOP_TAG, PERSIS_STOP,
                                         FINISHED_PERSISTENT_GEN_TAG,
                                         EVAL_GEN_TAG)
from libensemble.tools.persistent_support import PersistentSupport


def get_stage(persis_info):
    return str(persis_info['stage_count']).zfill(4)


def update_config_file(user, app_type, pinfo):
    """
    Updates configuration files for each application prior to launching, and
    produces expected output directory structure for each.
    """
    with open(user[app_type + '_config'], 'r') as f:
        config = yaml.safe_load(f)

    output_path = os.getcwd() + '/{}_runs/stage'.format(app_type) +  \
                                get_stage(pinfo) + '/task0000'

    updates = {
        'experiment_directory': os.getcwd(),
        'output_path': output_path,
        'stage_idx': pinfo['stage_count']
    }

    if app_type == 'aggregation':
        updates.update({'last_n_h5_files': user['initial_sample_size']})

    elif app_type == 'machine_learning':
        updates.update({
            'model_tag': 'keras_cvae_model' + get_stage(pinfo),
            'last_n_h5_files': user['last_n_h5_files']
        })

    elif app_type == 'model_selection':
        updates.update({'checkpoint_dir': output_path.replace(app_type, 'machine_learning') + '/checkpoint'})

    elif app_type == 'agent':
        updates.update({
            'num_intrinsic_outliers': user['outliers'],
            'num_extrinsic_outliers': user['outliers'],
            'n_most_recent_h5_files': user['n_most_recent_h5_files'],
            'n_traj_frames': user['n_traj_frames']
        })

    os.makedirs(output_path, exist_ok=True)
    task_config = os.path.join(output_path, 'stage' + get_stage(pinfo) + '_task0000.yaml')

    config.update(updates)

    with open(task_config, 'w') as f:
        yaml.dump(config, f)

    return output_path, task_config


def submit_application(exctr, user, app_type, output_path, task_config):
    """
    Switches to an expected output directory, launches an application
    via libEnsemble's executor, then polls its status until it finishes.
    """
    start = os.getcwd()
    os.chdir(output_path)

    args = '-c ' + os.path.join(os.getcwd(), task_config)
    task = exctr.submit(app_name=app_type, app_args=args, wait_on_start=True,
                        num_procs=1, num_nodes=1, procs_per_node=1)

    calc_status = exctr.polling_loop(task, timeout=user[app_type + '_kill_minutes']*60, delay=1)
    os.chdir(start)
    return calc_status


def postprocess_md_sim_dirs(calc_in, pinfo):
    """
    Symlink the Molecular Dynamics results into directories that resemble
    DeepDriveMD's output.
    """
    expected_md_dir = './molecular_dynamics_runs/stage' + get_stage(pinfo)
    os.makedirs(expected_md_dir)
    for entry in calc_in:
        base_task_dir = 'task' + str(entry['task_id']).zfill(4)
        full_task_dir = os.path.join(expected_md_dir, base_task_dir)
        sim_dir = entry['sim_dir_loc']
        os.symlink(os.path.abspath('../' + sim_dir), os.path.abspath(full_task_dir))


def generate_initial_md_runs(gen_specs, persis_info):
    """
    Generate an initial local History array, and populate with an initial set
    of parameters for an initial set of MD simulations.
    """
    persis_info['stage_count'] += 1

    sample_size = gen_specs['user']['initial_sample_size']
    local_H = np.zeros(sample_size, dtype=gen_specs['out'])

    local_H['task_id'] = np.arange(sample_size)
    local_H['initial'] = True
    local_H['gen_dir_loc'] = os.getcwd().split('/')[-1]
    local_H['sim_id'] = np.arange(sample_size)
    local_H['stage_id'] = 0

    persis_info['last_sim_id'] = local_H['sim_id'][-1]

    return local_H, persis_info


def generate_subsequent_md_runs(gen_specs, persis_info, local_H, output_path):
    """
    Generate subsequent MD simulation run parameters in the local History array,
    based on the number of outlier points detected by the Agent application.
    """
    persis_info['stage_count'] += 1

    presumed_agent_output = glob.glob(output_path + '/stage*_task*.json')[0]
    with open(os.path.join(output_path, presumed_agent_output), 'r') as f:
        sample_size = len(json.load(f))

    local_H.resize(len(local_H) + sample_size, refcheck=False)

    local_H['task_id'][-sample_size:] = np.arange(sample_size)
    local_H['initial'][-sample_size:] = False
    local_H['gen_dir_loc'][-sample_size:] = os.getcwd().split('/')[-1]

    subs_sim_id = persis_info['last_sim_id'] + 1
    local_H['sim_id'][-sample_size:] = np.arange(subs_sim_id, subs_sim_id + sample_size)
    local_H['stage_id'][-sample_size:] = persis_info['stage_count']

    persis_info['last_sim_id'] = local_H['sim_id'][-1]

    return local_H, persis_info


def skip_app(gen_specs, app):
    """
    Optionally skip certain apps, if specified in gen_specs['user']
    """
    if 'skip_' + app in gen_specs['user']:
        if gen_specs['user']['skip_' + app]:
            return True
    return False


def run_keras_cvae_ml_genf(H, persis_info, gen_specs, libE_info):
    """ Persistent Generator user function for processing simulation output and
    running via the Executor each of the remaining DeepDriveMD applications concerned
    with simulation output. This generator does not return until libEnsemble
    concludes.

    On initialization, this generator function produces an initial set of parameters
    for an initial set of simulation function calls, then sends the local History
    array containing these values directly to the Manager, which will distribute
    the work accordingly.

    After this, the persistent generator waits until all the results are available
    from the Manager, preprocesses some of the output, then configures and launches
    the other DeepDriveMD applications in a sequence. The final app's output
    determines the number of future candidate simulations. The local History array
    is updated, then sent directly to the Manager.
    """

    user = gen_specs['user']
    exctr = Executor.executor
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)
    persis_info['stage_count'] = -1
    os.environ["OMP_NUM_THREADS"] = '4'
    os.environ["CUDA_VISIBLE_DEVICES"] = str(libE_info['workerID'] - 1)
    initial_complete = False
    tag = None

    while True:
        if not initial_complete:
            local_H, persis_info = generate_initial_md_runs(gen_specs, persis_info)
            # Send initial MD run parameters directly to the Manager
            ps.send(local_H)
            initial_complete = True
        else:
            # Wait for batch of MD results
            tag, Work, calc_in = ps.recv()
            if tag in [STOP_TAG, PERSIS_STOP]:  # Generator instructed to stop
                break

            # Symlink MD data into directory structure expected by future apps
            postprocess_md_sim_dirs(calc_in, persis_info)

            # Run each subsequent DeepDriveMD app
            for app in ['aggregation', 'machine_learning', 'model_selection', 'agent']:
                if skip_app(gen_specs, app):
                    continue
                output_path, task_config = update_config_file(user, app, persis_info)
                calc_status = submit_application(exctr, user, app, output_path, task_config)
                local_H[app + '_cstat'][Work['libE_info']['H_rows']] = calc_status

            # Produce subsequent set of MD runs parameters based on the final app's results
            local_H, persis_info = generate_subsequent_md_runs(gen_specs, persis_info, local_H, output_path)
            # Send subsequent MD run parameters directly to the Manager
            ps.send(local_H)

    return local_H, persis_info, FINISHED_PERSISTENT_GEN_TAG



import logging

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


# unlesss resource sets is set explicitly. This class maps rsets to hardware resources.
class RSetResources():
    """A class that creates a fixed mapping of resource sets to the available resources.

    **Object Attributes:**

    These are set on initialisation and include inherited.
    ``rsets`` below is used to abbreviate ``resource sets``.

    :ivar int num_workers: Total number of workers
    :ivar int num_workers_2assign2: The number of workers that will be assigned resource sets.
    :ivar int total_num_rsets: The total number of resource sets.
    :ivar list split_list: A list of lists, where each element is the list of nodes for a given rset.
    :ivar list local_rsets_list: A list over rsets, where each element is the number of rsets that share the node.
    :ivar int rsets_per_node: The number of rsets per node (if an rset > 1 node, this will be 1)
    """

    def __init__(self, num_workers, resources):
        """Initializes a new RSetResources instance

        Determines the compute resources available for each resource set.

        Unless resource sets is set explicitly, the number of resource sets is the number of workers,
        excluding any workers defined as zero resource workers.

        Parameters
        ----------

        num_workers: int
            The total number of workers

        resources: Resources
            A Resources object containing global nodelist and intranode information

        """
        self.num_workers = num_workers
        self.num_workers_2assign2 = RSetResources.get_workers2assign2(self.num_workers, resources)
        self.total_num_rsets = resources.num_resource_sets or self.num_workers_2assign2
        self.split_list, self.local_rsets_list = \
            RSetResources.get_partitioned_nodelist(self.total_num_rsets, resources)
        self.rsets_per_node = RSetResources.get_rsets_on_a_node(self.total_num_rsets, resources)

    @staticmethod
    def best_split(a, n):
        """Creates the most even split of list a into n parts and return list of lists"""
        k, m = divmod(len(a), n)
        return (a[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n))

    @staticmethod
    def get_rsets_on_a_node(num_rsets, resources):
        """Returns the number of resource sets that can be placed on each node

        If there are more nodes than resource sets, returns 1.
        """
        num_nodes = len(resources.global_nodelist)
        # Round up if there's a remainder
        rsets_per_node = num_rsets//num_nodes + (num_rsets % num_nodes > 0)
        return rsets_per_node

    @staticmethod
    def get_workers2assign2(num_workers, resources):
        """Returns workers to assign resources to"""
        zero_resource_list = resources.zero_resource_workers
        return num_workers - len(zero_resource_list)

    @staticmethod
    def even_assignment(nnodes, nworkers):
        """Returns True if workers are evenly distributied to nodes, else False"""
        return nnodes % nworkers == 0 or nworkers % nnodes == 0

    @staticmethod
    def expand_list(nnodes, nworkers, nodelist):
        """Duplicates each element of ``nodelist`` to best map workers to nodes.

        Returns node list with duplicates, and a list of local (on-node) worker
        counts, both indexed by worker.
        """
        k, m = divmod(nworkers, nnodes)
        dup_list = []
        local_rsets_list = []
        for i, x in enumerate(nodelist):
            repeats = k + 1 if i < m else k
            for j in range(repeats):
                dup_list.append(x)
                local_rsets_list.append(repeats)
        return dup_list, local_rsets_list

    @staticmethod
    def get_split_list(num_rsets, resources):
        """Returns a list of lists for each worker

        Assumes that self.global_nodelist has been calculated (in __init__).
        """
        global_nodelist = resources.global_nodelist
        num_nodes = len(global_nodelist)

        if not RSetResources.even_assignment(num_nodes, num_rsets):
            logger.warning('Resource sets ({}) are not distributed evenly to available nodes ({})'
                           .format(num_rsets, num_nodes))

        # If multiple workers per node - create global node_list with N duplicates (for N workers per node)
        sub_node_workers = (num_rsets >= num_nodes)
        if sub_node_workers:
            global_nodelist, local_rsets_list = \
                RSetResources.expand_list(num_nodes, num_rsets, global_nodelist)
        else:
            local_rsets_list = [1] * num_rsets

        # Divide global list between workers
        split_list = list(RSetResources.best_split(global_nodelist, num_rsets))
        logger.debug("split_list is {}".format(split_list))
        return split_list, local_rsets_list

    @staticmethod
    def get_partitioned_nodelist(num_rsets, resources):
        """Returns lists of nodes available to all resource sets

        Assumes that self.global_nodelist has been calculated (in __init__).
        Also self.global_nodelist will have already removed non-application nodes
        """
        split_list, local_rsets_list = RSetResources.get_split_list(num_rsets, resources)
        return split_list, local_rsets_list



"""
Runs libEnsemble with the fd_param_finder persistent gen_f, which finds an
appropriate finite-difference parameter for the sim_f mapping from R^n to R^p
around the point x.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_fd_param_finder.py

When running with the above command, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi
# TESTSUITE_NPROCS: 4
# TESTSUITE_OS_SKIP: OSX
# TESTSUITE_EXTRA: true

import sys
import numpy as np
import shutil  # For ECnoise.m

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.noisy_vector_mapping import func_wrapper as sim_f, noisy_function
from libensemble.gen_funcs.persistent_fd_param_finder import fd_param_finder as gen_f
from libensemble.alloc_funcs.start_fd_persistent import finite_diff_alloc as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

x0 = np.array([1.23, 4.56])  # point about which we are calculating finite difference parameters
f0 = noisy_function(x0)
n = len(x0)
p = len(f0)

sim_specs = {
    'sim_f': sim_f,
    'in': ['x', 'f_ind'],
    'out': [('f_val', float)],
}

# The initial noise_h_mat is chosen to ECNoise both grows and shrinks the fd param
gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['x', 'f_val', 'n_ind', 'f_ind', 'x_ind', 'sim_id'],
    'out': [('x', float, (n,)), ('n_ind', int), ('f_ind', int), ('x_ind', int)],
    'user': {
        'x0': x0,
        'f0': f0,
        'nf': 10,
        'p': p,
        'n': n,
        'noise_h_mat': np.multiply(np.logspace(-16, -1, p), np.ones((n, p))),
        'maxnoiseits': 3,
    },
}
shutil.copy('./scripts_used_by_reg_tests/ECnoise.m', './')

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'gen_max': 1000}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert len(H) < exit_criteria['gen_max'], "Problem didn't stop early, which should have been the case."
    assert np.all(persis_info[1]['Fnoise'] > 0), "gen_f didn't find noise for all F_i components."

    save_libE_output(H, persis_info, __file__, nworkers)



import numpy as np

bounds = np.array([[63070, 115600],
                   [63.1, 116],
                   [990, 1110],
                   [700, 820],
                   [0, np.inf],  # Not sure if the physics have a more meaningful upper bound
                   [0.05, 0.15],  # Very low probability of being outside of this range
                   [9855, 12045],
                   [1120, 1680]])


def borehole(H, persis_info, sim_specs, _):
    """
    Wraps the borehole function
    """
    H_o = np.zeros(H['x'].shape[0], dtype=sim_specs['out'])
    H_o['f'] = borehole_func(H['x'])

    return H_o, persis_info


def borehole_func(x):
    """This evaluates the Borehole function for n-by-8 input
    matrix x, and returns the flow rate through the Borehole. (Harper and Gupta, 1983)
    input:

    Parameters
    ----------

    x: matrix of dimension (n, 8), where n is the number of input configurations:

        .. code-block::

        x[:,0]: Tu, transmissivity of upper aquifer (m^2/year)
        x[:,1]: Tl, transmissivity of lower aquifer (m^2/year)
        x[:,2]: Hu, potentiometric head of upper aquifer (m)
        x[:,3]: Hl, potentiometric head of lower aquifer (m)
        x[:,4]: r, radius of influence (m)
        x[:,5]: rw, radius of borehole (m)
        x[:,6]: Kw, hydraulic conductivity of borehole (m/year)
        x[:,7]: L, length of borehole (m)

    Returns
    -------

    vector of dimension (n, 1):
        flow rate through the Borehole (m^3/year)

    """

    assert np.all(x >= bounds[:, 0]) and np.all(x <= bounds[:, 1]), "Point not within bounds"

    axis = 1
    if x.ndim == 1:
        axis = 0

    (Tu, Tl, Hu, Hl, r, rw, Kw, L) = np.split(x, 8, axis)

    numer = 2 * np.pi * Tu * (Hu - Hl)
    denom1 = 2 * L * Tu / (np.log(r/rw) * rw**2 * Kw)
    denom2 = Tu / Tl

    return (numer / (np.log(r/rw) * (1 + denom1 + denom2))).reshape(-1)


def gen_borehole_input(n):
    """Generates and returns n inputs for the Borehole function, according to distributions
    outlined in Harper and Gupta (1983).

    input:
      n: number of input to generate
    output:
      matrix of (n, 8), input to borehole_func(x) function
    """

    Tu = np.random.uniform(63070, 115600, n)
    Tl = np.random.uniform(63.1, 116, n)
    Hu = np.random.uniform(990, 1110, n)
    Hl = np.random.uniform(700, 820, n)
    r = np.random.lognormal(7.71, 1.0056, n)
    rw = np.random.normal(0.1, 0.0161812, n)
    Kw = np.random.uniform(9855, 12045, n)
    L = np.random.uniform(1120, 1680, n)

    x = np.column_stack((Tu, Tl, Hu, Hl, r, rw, Kw, L))
    return x



"""
This module contains methods used our implementation of the Asynchronously
Parallel Optimization Solver for finding Multiple Minima (APOSMM) method
described in detail in :cite:`LW16`.
"""
__all__ = ['aposmm_logic', 'initialize_APOSMM',
           'decide_where_to_start_localopt', 'update_history_dist']

import sys
import pickle
import traceback
import numpy as np
from scipy.spatial.distance import cdist, pdist, squareform
from numpy.lib.recfunctions import merge_arrays
from math import log, gamma, pi, sqrt

import libensemble.gen_funcs
optimizer_list = ['petsc', 'nlopt', 'scipy']
optimizers = libensemble.gen_funcs.rc.aposmm_optimizers

if optimizers is None:
    from petsc4py import PETSc
    import nlopt
    from scipy import optimize as scipy_optimize
else:
    if not isinstance(optimizers, list):
        optimizers = [optimizers]
    unrec = set(optimizers) - set(optimizer_list)
    if unrec:
        print('APOSMM Warning: unrecognized optimizers {}'.format(unrec))

    if 'petsc' in optimizers:
        from petsc4py import PETSc
    if 'nlopt' in optimizers:
        import nlopt
    if 'scipy' in optimizers:
        from scipy import optimize as scipy_optimize


class APOSMMException(Exception):
    "Raised for any exception in APOSMM"


def aposmm_logic(H, persis_info, gen_specs, _):
    """
    APOSMM coordinates multiple local optimization runs, starting from points
    which do not have a better point nearby (within a distance ``r_k``). This
    generation function produces/requires the following fields in ``H``:

    - ``'x' [n floats]``: Parameters being optimized over
    - ``'x_on_cube' [n floats]``: Parameters scaled to the unit cube
    - ``'f' [float]``: Objective function being minimized
    - ``'local_pt' [bool]``: True if point from a local optimization run
    - ``'dist_to_unit_bounds' [float]``: Distance to domain boundary
    - ``'dist_to_better_l' [float]``: Dist to closest better local opt point
    - ``'dist_to_better_s' [float]``: Dist to closest better sample point
    - ``'ind_of_better_l' [int]``: Index of point ``'dist_to_better_l``' away
    - ``'ind_of_better_s' [int]``: Index of point ``'dist_to_better_s``' away
    - ``'started_run' [bool]``: True if point has started a local opt run
    - ``'num_active_runs' [int]``: Number of active local runs point is in
    - ``'local_min' [float]``: True if point has been ruled a local minima
    - ``'sim_id' [int]``: Row number of entry in history

    and optionally

    - ``'priority' [float]``: Value quantifying a point's desirability
    - ``'f_i' [float]``: Value of ith objective component (if single_component)
    - ``'fvec' [m floats]``: All objective components (if calculated together)
    - ``'obj_component' [int]``: Index corresponding to value in ``'f_i``'
    - ``'pt_id' [int]``: Identify the point (useful when evaluating different
      objective components for a given ``'x'``)

    When using libEnsemble to do individual objective component evaluations,
    APOSMM will return ``gen_specs['user']['components']`` copies of each point, but
    the component=0 entry of each point will be considered only when

    - deciding where to start a run,
    - determining the best nearby point,
    - storing the order of the points in the run, or
    - storing the combined objective function value

    Necessary quantities in ``gen_specs['user']`` are

    - ``'lb' [n floats]``: Lower bound on search domain
    - ``'ub' [n floats]``: Upper bound on search domain
    - ``'initial_sample_size' [int]``: Number of uniformly sampled points
      must be returned (non-nan value) before a local opt run is started

    - ``'localopt_method' [str]``: Name of an NLopt, PETSc/TAO, or SciPy method
      (see 'advance_local_run' below for supported methods)

    Optional ``gen_specs['user']`` entries are as follows

    - ``'sample_points' [numpy array]``: Points to be sampled (original domain)
    - ``'combine_component_func' [func]``: Function to combine obj components
    - ``'components' [int]``: Number of objective components
    - ``'dist_to_bound_multiple' [float in (0,1]]``: Fraction of the
      distance to the nearest boundary to be used for the initial step size be
      in localopt runs
    - ``'high_priority_to_best_localopt_runs': [bool]``: True if localopt runs
      with smallest observed function value are given priority
    - ``'lhs_divisions' [int]``: Number of Latin hypercube sampling partitions
      (0 or 1 results in uniform sampling)
    - ``'min_batch_size' [int]``: Lower bound on the number of points given
      every time APOSMM is called
    - ``'mu' [float]``: Distance from the boundary that all localopt starting
      points must satisfy
    - ``'nu' [float]``: Distance from identified minima that all starting
      points must satisfy
    - ``'single_component_at_a_time' [bool]``: True if single objective
      components will be evaluated at a time
    - ``'rk_const' [float]``: Multiplier in front of the r_k value
    - ``'max_active_runs' [int]``: Bound on number of runs APOSMM is advancing

    The following are ``gen_specs['user']`` convergence tolerances for NLopt, PETSc/TAO, SciPy

    - ``'fatol' [float]``:
    - ``'ftol_abs' [float]``:
    - ``'ftol_rel' [float]``:
    - ``'gatol' [float]``:
    - ``'grtol' [float]``:
    - ``'xtol_abs' [float]``:
    - ``'xtol_rel' [float]``:
    - ``'tol' [float]``:


    As a default, APOSMM starts a local optimization run from a point that

    - is not in an active local optimization run,
    - is more than ``mu`` from the boundary (in the unit-cube domain),
    - is more than ``nu`` from identified minima (in the unit-cube domain), and
    - does not have a better point within a distance ``r_k`` of it.

    If the above results in more than ``'max_active_runs'`` being advanced, the
    best point in each run is determined, and the dist_to_better is computed
    (with inf being the value for the best run). Then those
    ``'max_active_runs'`` runs with largest dist_to_better are advanced
    (breaking ties arbitrarily).

    :Note:
        ``gen_specs['user']['combine_component_func']`` must be defined when there are
        multiple objective components.

    :Note:
        APOSMM critically uses ``persis_info`` to store information about
        active runs, order of points in each run, etc. The allocation function
        must ensure that it is always given.

    .. seealso::
        `test_sim_dirs.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_sim_dirs.py>`_
        for basic APOSMM usage.

    .. seealso::
        `test_old_aposmm_one_residual_at_a_time.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_old_aposmm_one_residual_at_a_time.py>`_
        for an example of APOSMM coordinating multiple local optimization runs
        for an objective with more than one component.
    """
    """
    Description of intermediate variables in aposmm_logic:

    n:                domain dimension
    c_flag:           True if giving libEnsemble individual components of fvec
                      to evaluate. (Note if c_flag is True, APOSMM will use
                      only the component to store the function value f)
    n_s:              the number of complete evaluations of sampled points
    updated_inds:     indices of H that have been updated (and so all their
                      information must be sent back to libE manager to update)
    H_o:                new points to be sent back to the history


    When re-running a local opt method to get the next point:
    advance_local_run.x_new:      stores the first new point requested by
                                  a local optimization method
    advance_local_run.pt_in_run:  counts function evaluations to know
                                  when a new point is given

    starting_inds:    indices where a runs should be started.
    active_runs:      indices of active local optimization runs
    sorted_run_inds:  indices of the considered run (in the order they were
                      requested by the localopt method)
    x_opt:            the reported minimum from a localopt run (disregarded
                      unless exit_code isn't 0)
    exit_code:        0 if a new localopt point has been found, otherwise it's
                      the NLopt/TAO/SciPy code
    samples_needed:   Number of additional uniformly drawn samples needed


    Description of persistent variables used to maintain the state of APOSMM

    persis_info['total_runs']: Running count of started/completed localopt runs
    persis_info['run_order']: Sequence of indices of points in unfinished runs
    persis_info['old_runs']: Sequence of indices of points in finished runs

    """

    n, n_s, c_flag, H_o, r_k, mu, nu = initialize_APOSMM(H, gen_specs)

    # np.savez('H'+str(len(H)),H=H,gen_specs=gen_specs,persis_info=persis_info)
    if n_s < gen_specs['user']['initial_sample_size']:
        updated_inds = set()

    else:
        updated_inds = update_history_dist(H, n, gen_specs, c_flag)

        starting_inds = decide_where_to_start_localopt(H, r_k, mu, nu)
        updated_inds.update(starting_inds)

        for ind in starting_inds:
            # Find the run number
            new_run_num = persis_info['total_runs']

            H['started_run'][ind] = 1
            H['num_active_runs'][ind] += 1

            persis_info['run_order'][new_run_num] = [ind]
            persis_info['total_runs'] += 1

        num_runs = len(persis_info['run_order'])
        if 'max_active_runs' in gen_specs['user'] and gen_specs['user']['max_active_runs'] < num_runs:
            # Store run number and sim_id of the best point in each run
            run_vals = np.zeros((num_runs, 2), dtype=int)
            for i, run in enumerate(persis_info['run_order'].keys()):
                run_vals[i, 0] = run
                run_vals[i, 1] = persis_info['run_order'][run][np.nanargmin(
                    H['f'][persis_info['run_order'][run]])]

            # Compute pairwise distance between the best points in each run
            P = squareform(pdist(H['x_on_cube'][run_vals[:, 1]], 'euclidean'))
            dist_to_better = np.inf*np.ones(num_runs)

            for i in range(num_runs):
                better = H['f'][run_vals[:, 1]] < H['f'][run_vals[i, 1]]
                if any(better):
                    dist_to_better[i] = np.min(P[i, better])

            # Take max_active_runs largest
            k_sorted = np.argpartition(-dist_to_better, kth=gen_specs['user']['max_active_runs']-1)
            active_runs = set(run_vals[k_sorted[:gen_specs['user']['max_active_runs']], 0].astype(int))
        else:
            active_runs = set(persis_info['run_order'].keys())

        inactive_runs = set()

        # Find next point in any uncompleted run using persis_info['run_order']
        for run in active_runs:
            if not np.all(H['returned'][persis_info['run_order'][run]]):
                continue  # Can't advance a run if all points aren't returned.

            x_opt, exit_code, persis_info, sorted_run_inds, x_new = advance_local_run(H, gen_specs['user'], c_flag, run, persis_info)

            if np.isinf(x_new).all():
                if exit_code == 0:
                    run_out_file = "run_" + str(run) + "_abort.pickle"

                    with open(run_out_file, "wb") as f:
                        pickle.dump((H, gen_specs, c_flag, run, persis_info), f)

                    raise APOSMMException("Exit code is 0, but x_new was not updated in " +
                                          "local opt run " + str(run) + " after " +
                                          str(len(sorted_run_inds)) + " evaluations.\n" +
                                          "Saving run information to: " + run_out_file +
                                          "\nWorker crashing!")

                # No new point was added. Hopefully at a minimum
                update_history_optimal(x_opt, H, sorted_run_inds)
                inactive_runs.add(run)
                updated_inds.update(sorted_run_inds)

            else:
                # Check if x_new is already being requested (a check if it's in
                # H is performed inside advance_local_run)
                match_ind = np.where(np.equal(x_new, H_o['x_on_cube']).all(1))[0]
                if len(match_ind) == 0:
                    persis_info = add_to_Out(H_o, x_new, H, gen_specs, c_flag,
                                             persis_info, local_flag=1,
                                             sorted_run_inds=sorted_run_inds,
                                             run=run)
                else:
                    assert len(match_ind) == 1, "The same point is in H_o twice"
                    persis_info['run_order'][run].append(H_o['sim_id'][match_ind[0]])

        for i in inactive_runs:
            old_run = persis_info['run_order'].pop(i)  # Deletes all run info
            persis_info['old_runs'][i] = old_run

    if len(H) == 0:
        samples_needed = gen_specs['user']['initial_sample_size']
    elif 'min_batch_size' in gen_specs['user']:
        samples_needed = gen_specs['user']['min_batch_size']-len(H_o)
    else:
        samples_needed = int(not bool(len(H_o)))  # 1 if len(H_o)==0, 0 otherwise

    if samples_needed > 0 and 'sample_points' in gen_specs['user']:
        v = np.sum(~H['local_pt'])  # Number of sample points so far
        sampled_points = gen_specs['user']['sample_points'][v:v+samples_needed]
        on_cube = False  # Assume points are on original domain, not unit cube
        if len(sampled_points):
            persis_info = add_to_Out(H_o, sampled_points, H, gen_specs,
                                     c_flag, persis_info, on_cube=on_cube)
        samples_needed = samples_needed-len(sampled_points)

    if samples_needed > 0:
        sampled_points = persis_info['rand_stream'].uniform(0, 1, (samples_needed, n))
        on_cube = True
        persis_info = add_to_Out(H_o, sampled_points, H, gen_specs, c_flag,
                                 persis_info, on_cube=on_cube)

    H_o = np.append(H[np.array(list(updated_inds), dtype=int)][[o[0] for o in gen_specs['out']]], H_o)

    return H_o, persis_info


def add_to_Out(H_o, pts, H, gen_specs, c_flag, persis_info, local_flag=0,
               sorted_run_inds=[], run=[], on_cube=True):
    """
    Adds points to H_o, the numpy structured array to be sent back to the manager
    """

    assert not local_flag or len(pts) == 1, "Can't > 1 local points"

    original_len_O = len(H_o)

    len_H = len(H)
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    if c_flag:
        m = gen_specs['user']['components']

        assert len_H % m == 0, "Number of points in len_H not congruent to 0 mod 'components'"
        pt_ids = np.sort(np.tile(np.arange((len_H+original_len_O)/m, (len_H+original_len_O)/m+len(pts)), (1, m)))
        pts = np.tile(pts, (m, 1))

    num_pts = len(pts)

    H_o.resize(len(H_o)+num_pts, refcheck=False)  # Adds num_pts rows of zeros to H_o

    if on_cube:
        H_o['x_on_cube'][-num_pts:] = pts
        H_o['x'][-num_pts:] = pts*(ub-lb)+lb
    else:
        H_o['x_on_cube'][-num_pts:] = (pts-lb)/(ub-lb)
        H_o['x'][-num_pts:] = pts

    H_o['sim_id'][-num_pts:] = np.arange(len_H+original_len_O, len_H+original_len_O+num_pts)
    H_o['local_pt'][-num_pts:] = local_flag

    H_o['dist_to_unit_bounds'][-num_pts:] = np.inf
    H_o['dist_to_better_l'][-num_pts:] = np.inf
    H_o['dist_to_better_s'][-num_pts:] = np.inf
    H_o['ind_of_better_l'][-num_pts:] = -1
    H_o['ind_of_better_s'][-num_pts:] = -1

    if c_flag:
        H_o['obj_component'][-num_pts:] = np.tile(range(0, m), (1, num_pts//m))
        H_o['pt_id'][-num_pts:] = pt_ids

    if local_flag:
        H_o['num_active_runs'][-num_pts] += 1
        # H_o['priority'][-num_pts:] = 1
        # H_o['priority'][-num_pts:] = np.random.uniform(0,1,num_pts)
        if 'high_priority_to_best_localopt_runs' in gen_specs['user'] and gen_specs['user']['high_priority_to_best_localopt_runs']:
            H_o['priority'][-num_pts:] = -min(H['f'][persis_info['run_order'][run]])  # Give highest priority to run with lowest function value
        else:
            H_o['priority'][-num_pts:] = persis_info['rand_stream'].uniform(0, 1, num_pts)
        persis_info['run_order'][run].append(H_o[-num_pts]['sim_id'])
    else:
        if c_flag:
            # p_tmp = np.sort(np.tile(np.random.uniform(0,1,num_pts/m),(m,1))) # If you want all "duplicate points" to have the same priority (meaning libEnsemble gives them all at once)
            # p_tmp = np.random.uniform(0,1,num_pts)
            p_tmp = persis_info['rand_stream'].uniform(0, 1, num_pts)
        else:
            # p_tmp = np.random.uniform(0,1,num_pts)
            # persis_info['rand_stream'].uniform(lb,ub,(1,n))
            if 'high_priority_to_best_localopt_runs' in gen_specs['user'] and gen_specs['user']['high_priority_to_best_localopt_runs']:
                p_tmp = -np.inf*np.ones(num_pts)
            else:
                p_tmp = persis_info['rand_stream'].uniform(0, 1, num_pts)
        H_o['priority'][-num_pts:] = p_tmp
        # H_o['priority'][-num_pts:] = 1

    return persis_info


def update_history_dist(H, n, gen_specs, c_flag):
    """
    Updates distances/indices after new points that have been evaluated.

    .. seealso::
        `start_persistent_local_opt_gens.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/alloc_funcs/start_persistent_local_opt_gens.py>`_
    """

    updated_inds = set()

    new_inds = np.where(~H['known_to_aposmm'])[0]

    if c_flag:
        for v in np.unique(H['pt_id'][new_inds]):
            inds = H['pt_id'] == v
            H['f'][inds] = np.inf
            H['f'][np.where(inds)[0][0]] = gen_specs['user']['combine_component_func'](H['f_i'][inds])

        p = np.logical_and.reduce((H['returned'], H['obj_component'] == 0, ~np.isnan(H['f'])))
    else:
        p = np.logical_and.reduce((H['returned'], ~np.isnan(H['f'])))

    for new_ind in new_inds:
        # Loop over new returned points and update their distances
        if p[new_ind]:
            H['known_to_aposmm'][new_ind] = True

            # Compute distance to boundary
            H['dist_to_unit_bounds'][new_ind] = min(min(np.ones(n)-H['x_on_cube'][new_ind]), min(H['x_on_cube'][new_ind]-np.zeros(n)))

            dist_to_all = cdist(H['x_on_cube'][[new_ind]], H['x_on_cube'][p], 'euclidean').flatten()
            new_better_than = H['f'][new_ind] < H['f'][p]

            # Update any other points if new_ind is closer and better
            if H['local_pt'][new_ind]:
                inds_of_p = np.logical_and(dist_to_all < H['dist_to_better_l'][p], new_better_than)
                updates = np.where(p)[0][inds_of_p]
                H['dist_to_better_l'][updates] = dist_to_all[inds_of_p]
                H['ind_of_better_l'][updates] = new_ind
            else:
                inds_of_p = np.logical_and(dist_to_all < H['dist_to_better_s'][p], new_better_than)
                updates = np.where(p)[0][inds_of_p]
                H['dist_to_better_s'][updates] = dist_to_all[inds_of_p]
                H['ind_of_better_s'][updates] = new_ind
            updated_inds.update(updates)

            # Since we allow equality when deciding better_than_new_l and
            # better_than_new_s, we have to prevent new_ind from being its own
            # better point.
            better_than_new_l = np.logical_and.reduce((~new_better_than, H['local_pt'][p], H['sim_id'][p] != new_ind))
            better_than_new_s = np.logical_and.reduce((~new_better_than, ~H['local_pt'][p], H['sim_id'][p] != new_ind))

            # Who is closest to ind and better
            if np.any(better_than_new_l):
                ind = dist_to_all[better_than_new_l].argmin()
                H['ind_of_better_l'][new_ind] = H['sim_id'][p][np.nonzero(better_than_new_l)[0][ind]]
                H['dist_to_better_l'][new_ind] = dist_to_all[better_than_new_l][ind]

            if np.any(better_than_new_s):
                ind = dist_to_all[better_than_new_s].argmin()
                H['ind_of_better_s'][new_ind] = H['sim_id'][p][np.nonzero(better_than_new_s)[0][ind]]
                H['dist_to_better_s'][new_ind] = dist_to_all[better_than_new_s][ind]

            # if not ignore_L8:
            #     r_k = calc_rk(len(H['x_on_cube'][0]), n_s, rk_const, lhs_divisions)
            #     H['worse_within_rk'][new_ind][p] = np.logical_and.reduce((H['f'][new_ind] <= H['f'][p], dist_to_all <= r_k))

            #     # Add trues if new point is 'worse_within_rk'
            #     inds_to_change = np.logical_and.reduce((H['dist_to_all'][p,new_ind] <= r_k, H['f'][new_ind] >= H['f'][p], H['sim_id'][p] != new_ind))
            #     H['worse_within_rk'][inds_to_change,new_ind] = True

            #     if not H['local_pt'][new_ind]:
            #         H['worse_within_rk'][H['dist_to_all'] > r_k] = False

    updated_inds.update(new_inds)
    return updated_inds


def update_history_optimal(x_opt, H, run_inds):
    """
    Updated the history after any point has been declared a local minimum
    """

    # opt_ind = np.where(np.logical_and(np.equal(x_opt,H['x_on_cube']).all(1),~np.isinf(H['f'])))[0] # This fails on some problems. x_opt is 1e-16 away from the point that was given and opt_ind is empty
    run_inds = np.unique(run_inds)

    dists = np.linalg.norm(H['x_on_cube'][run_inds]-x_opt, axis=1)
    ind = np.argmin(dists)
    opt_ind = run_inds[ind]

    if dists[ind] > 1e-15:
        print("Dist from x_opt to closest point is:"+str(dists[ind]))
        print("Report this!")
        print(x_opt)
        print(run_inds, flush=True)
    assert dists[ind] <= 1e-15, "Closest point to x_opt not within 1e-15?"

    failsafe = np.logical_and(H['f'][run_inds] < H['f'][opt_ind], dists < 1e-8)
    if np.any(failsafe):
        # Rare event, but want to not start another run next to a minimum
        print('Marking more than 1 point in this run as a min!')
        print("Report this!", flush=True)
        H['local_min'][run_inds[failsafe]] = 1

    H['local_min'][opt_ind] = 1
    H['num_active_runs'][run_inds] -= 1


def advance_local_run(H, user_specs, c_flag, run, persis_info):
    """
    Moves a local optimization method one iteration forward. We currently do
    this by feeding all past evaluations from a run to the method and then
    storing the first new point generated
    """

    while 1:
        sorted_run_inds = persis_info['run_order'][run]
        advance_local_run.x_new = np.ones((1, len(user_specs['ub'])))*np.inf
        advance_local_run.pt_in_run = 0

        if user_specs['localopt_method'] in ['LN_SBPLX', 'LN_BOBYQA',
                                             'LN_COBYLA', 'LN_NELDERMEAD',
                                             'LD_MMA']:

            if user_specs['localopt_method'] in ['LD_MMA']:
                fields_to_pass = ['x_on_cube', 'f', 'grad']
            else:
                fields_to_pass = ['x_on_cube', 'f']

            try:
                x_opt, exit_code = set_up_and_run_nlopt(H[fields_to_pass][sorted_run_inds], user_specs)
            except Exception as e:
                x_opt = 0
                exit_code = 0
                display_exception(e)

        elif user_specs['localopt_method'] in ['pounders', 'blmvm']:

            if c_flag:
                Run_H_F = np.zeros(len(sorted_run_inds), dtype=[('fvec', float, user_specs['components'])])
                for i, ind in enumerate(sorted_run_inds):
                    a1 = H['pt_id'] == H['pt_id'][ind]
                    Run_H_F['fvec'][i, :] = H['f_i'][a1]
                Run_H = merge_arrays([H[['x_on_cube']][sorted_run_inds], Run_H_F], flatten=True)
            else:
                if user_specs['localopt_method'] == 'pounders':
                    Run_H = H[['x_on_cube', 'fvec']][sorted_run_inds]
                else:
                    Run_H = H[['x_on_cube', 'f', 'grad']][sorted_run_inds]

            try:
                x_opt, exit_code = set_up_and_run_tao(Run_H, user_specs)
            except Exception as e:
                x_opt = 0
                exit_code = 0
                display_exception(e)

        elif user_specs['localopt_method'] == 'scipy_COBYLA':

            fields_to_pass = ['x_on_cube', 'f']

            try:
                x_opt, exit_code = set_up_and_run_scipy_minimize(H[fields_to_pass][sorted_run_inds], user_specs)
            except Exception as e:
                x_opt = 0
                exit_code = 0
                display_exception(e)

        else:
            raise APOSMMException("Unknown localopt method. Exiting")

        match_ind = np.equal(advance_local_run.x_new, H['x_on_cube']).all(1)
        if ~match_ind.any():
            # Generated a new point
            break
        else:
            # We need to add a previously evaluated point into this run
            persis_info['run_order'][run].append(np.nonzero(match_ind)[0][0])

    return x_opt, exit_code, persis_info, sorted_run_inds, advance_local_run.x_new


def set_up_and_run_scipy_minimize(Run_H, user_specs):
    """ Set up objective and runs scipy

    Declares the appropriate syntax for our special objective function to read
    through Run_H, sets the parameters and starting points for the run.
    """

    def scipy_obj_fun(x, Run_H):
        out = look_in_history(x, Run_H)

        return out

    x0 = Run_H['x_on_cube'][0]

    # Construct the bounds in the form of constraints
    cons = []
    for factor in range(len(x0)):
        lo = {'type': 'ineq',
              'fun': lambda x, lb=user_specs['lb'][factor], i=factor: x[i]-lb}
        up = {'type': 'ineq',
              'fun': lambda x, ub=user_specs['ub'][factor], i=factor: ub-x[i]}
        cons.append(lo)
        cons.append(up)

    method = user_specs['localopt_method'][6:]
    res = scipy_optimize.minimize(lambda x: scipy_obj_fun(x, Run_H), x0, method=method, options={'maxiter': len(Run_H['x_on_cube'])+1, 'tol': user_specs['tol']})

    if res['status'] == 2:  # SciPy code for exhausting budget of evaluations, so not at a minimum
        exit_code = 0
    else:
        if method == 'COBYLA':
            assert res['status'] == 1, "Unknown status for COBYLA"
            exit_code = 1

    x_opt = res['x']
    return x_opt, exit_code


def set_up_and_run_nlopt(Run_H, user_specs):
    """ Set up objective and runs nlopt

    Declares the appropriate syntax for our special objective function to read
    through Run_H, sets the parameters and starting points for the run.
    """

    assert 'xtol_rel' or 'xtol_abs' or 'ftol_rel' or 'ftol_abs' in user_specs, "NLopt can cycle if xtol_rel, xtol_abs, ftol_rel, or ftol_abs are not set"

    def nlopt_obj_fun(x, grad, Run_H):
        out = look_in_history(x, Run_H)

        if user_specs['localopt_method'] in ['LD_MMA']:
            grad[:] = out[1]
            out = out[0]

        return out

    n = len(user_specs['ub'])

    opt = nlopt.opt(getattr(nlopt, user_specs['localopt_method']), n)

    lb = np.zeros(n)
    ub = np.ones(n)
    opt.set_lower_bounds(lb)
    opt.set_upper_bounds(ub)
    x0 = Run_H['x_on_cube'][0]

    # Care must be taken here because a too-large initial step causes nlopt to move the starting point!
    dist_to_bound = min(min(ub-x0), min(x0-lb))
    assert dist_to_bound > np.finfo(np.float64).eps, "The distance to the boundary is too small for NLopt to handle"

    if 'dist_to_bound_multiple' in user_specs:
        opt.set_initial_step(dist_to_bound*user_specs['dist_to_bound_multiple'])
    else:
        opt.set_initial_step(dist_to_bound)

    opt.set_maxeval(len(Run_H)+1)  # evaluate one more point
    opt.set_min_objective(lambda x, grad: nlopt_obj_fun(x, grad, Run_H))
    if 'xtol_rel' in user_specs:
        opt.set_xtol_rel(user_specs['xtol_rel'])
    if 'ftol_rel' in user_specs:
        opt.set_ftol_rel(user_specs['ftol_rel'])
    if 'xtol_abs' in user_specs:
        opt.set_xtol_abs(user_specs['xtol_abs'])
    if 'ftol_abs' in user_specs:
        opt.set_ftol_abs(user_specs['ftol_abs'])

    x_opt = opt.optimize(x0)
    exit_code = opt.last_optimize_result()

    if exit_code == 5:  # NLOPT code for exhausting budget of evaluations, so not at a minimum
        exit_code = 0

    return x_opt, exit_code


def set_up_and_run_tao(Run_H, user_specs):
    """ Set up objective and runs PETSc on the comm_self communicator

    Declares the appropriate syntax for our special objective function to read
    through Run_H, sets the parameters and starting points for the run.
    """
    tao_comm = PETSc.COMM_SELF
    n = len(user_specs['ub'])

    def pounders_obj_func(tao, X, F, Run_H):
        F.array = look_in_history(X.array_r, Run_H, vector_return=True)
        return F

    def blmvm_obj_func(tao, X, G, Run_H):
        (f, grad) = look_in_history(X.array_r, Run_H)
        G.array = grad
        return f

    # Create starting point, bounds, and tao object
    x = PETSc.Vec().create(tao_comm)
    x.setSizes(n)
    x.setFromOptions()
    x.array = Run_H['x_on_cube'][0]
    lb = x.duplicate()
    ub = x.duplicate()
    lb.array = 0*np.ones(n)
    ub.array = 1*np.ones(n)
    tao = PETSc.TAO().create(tao_comm)
    tao.setType(user_specs['localopt_method'])

    if user_specs['localopt_method'] == 'pounders':
        f = PETSc.Vec().create(tao_comm)
        f.setSizes(len(Run_H['fvec'][0]))
        f.setFromOptions()

        delta_0 = user_specs['dist_to_bound_multiple']*np.min([np.min(ub.array-x.array), np.min(x.array-lb.array)])

        PETSc.Options().setValue('-tao_pounders_delta', str(delta_0))

        # PETSc.Options().setValue('-pounders_subsolver_tao_type','bqpip')
        if hasattr(tao, 'setResidual'):
            tao.setResidual(lambda tao, x, f: pounders_obj_func(tao, x, f, Run_H), f)
        else:
            tao.setSeparableObjective(lambda tao, x, f: pounders_obj_func(tao, x, f, Run_H), f)

    elif user_specs['localopt_method'] == 'blmvm':
        g = PETSc.Vec().create(tao_comm)
        g.setSizes(n)
        g.setFromOptions()
        tao.setObjectiveGradient(lambda tao, x, g: blmvm_obj_func(tao, x, g, Run_H))

    # Set everything for tao before solving
    PETSc.Options().setValue('-tao_max_funcs', str(len(Run_H)+1))
    tao.setFromOptions()
    tao.setVariableBounds((lb, ub))
    # tao.setObjectiveTolerances(fatol=user_specs['fatol'], frtol=user_specs['frtol'])
    # tao.setGradientTolerances(grtol=user_specs['grtol'], gatol=user_specs['gatol'])
    tao.setTolerances(grtol=user_specs['grtol'], gatol=user_specs['gatol'])
    tao.setInitial(x)

    tao.solve(x)

    x_opt = tao.getSolution().getArray()
    exit_code = tao.getConvergedReason()
    # print(exit_code)
    # print(tao.view())
    # print(x_opt)

    if user_specs['localopt_method'] == 'pounders':
        f.destroy()
    if user_specs['localopt_method'] == 'blmvm':
        g.destroy()

    lb.destroy()
    ub.destroy()
    x.destroy()
    tao.destroy()

    return x_opt, exit_code


def decide_where_to_start_localopt(H, r_k, mu=0, nu=0, gamma_quantile=1):
    """
    Finds points in the history that satisfy the conditions (S1-S5 and L1-L8) in
    Table 1 of the `APOSMM paper <https://doi.org/10.1007/s12532-017-0131-4>`_
    This method first identifies sample points satisfying S2-S5 and then
    identifies all localopt points that satisfy L1-L7.
    We then start from any sample point also satisfying S1.
    We do not check condition L8 currently.

    We don't consider points in the history that have not returned from
    computation or that have a ``nan`` value. Also, note that ``mu`` and ``nu``
    implicitly depend on the scaling that is happening with the domain. That
    is, adjusting the initial domain can make a run start (or not start) at
    a point that didn't (or did) previously.

    Parameters
    ----------
    H: numpy structured array
        History array storing rows for each point.
    r_k_const: float
        Radius for deciding when to start runs
    lhs_divisions: integer
        Number of Latin hypercube sampling divisions (0 or 1 means uniform
        random sampling over the domain)
    mu: nonnegative float
        Distance from the boundary that all starting points must satisfy
    nu: nonnegative float
        Distance from identified minima that all starting points must satisfy
    gamma_quantile: float in (0,1]
        Only sample points whose function values are in the lower
        gamma_quantile can start localopt runs

    Returns
    ----------
    start_inds: list
        Indices where a local opt run should be started


    .. seealso::
        `start_persistent_local_opt_gens.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/alloc_funcs/start_persistent_local_opt_gens.py>`_
    """

    if nu > 0:
        test_2_through_5 = np.logical_and.reduce((
            H['returned'] == 1,  # have a returned function value
            H['dist_to_better_s'] >
            r_k,  # no better sample point within r_k (L2)
            ~H['started_run'],  # have not started a run (L3)
            H['dist_to_unit_bounds'] >=
            mu,  # have all components at least mu away from bounds (L4)
            np.all(
                cdist(H['x_on_cube'], H['x_on_cube'][H['local_min']]) >= nu,
                axis=1)  # distance nu away from known local mins (L5)
        ))
    else:
        test_2_through_5 = np.logical_and.reduce((
            H['returned'] == 1,  # have a returned function value
            H['dist_to_better_s'] >
            r_k,  # no better sample point within r_k (L2)
            ~H['started_run'],  # have not started a run (L3)
            H['dist_to_unit_bounds'] >=
            mu,  # have all components at least mu away from bounds (L4)
        ))  # (L5) is always true when nu = 0

    assert gamma_quantile == 1, "This is not supported yet. What is the best way to decide this when there are NaNs present in H['f']?"
    # if gamma_quantile < 1:
    #     cut_off_value = np.sort(H['f'][~H['local_pt']])[np.floor(gamma_quantile*(sum(~H['local_pt'])-1)).astype(int)]
    # else:
    #     cut_off_value = np.inf

    # Find the indices of points that...
    sample_seeds = np.logical_and.reduce((
        ~H['local_pt'],  # are not localopt points
        # H['f'] <= cut_off_value,      # have a small enough objective value
        ~np.isinf(H['f']),  # have a non-infinity objective value
        ~np.isnan(H['f']),  # have a non-NaN objective value
        test_2_through_5,  # satisfy tests 2 through 5
    ))

    # Uncomment the following to test the effect of ignoring LocalOpt points
    # in APOSMM. This allows us to test a parallel MLSL.
    # return list(np.ix_(sample_seeds)[0])

    those_satisfying_S1 = H['dist_to_better_l'][sample_seeds] > r_k  # no better localopt point within r_k
    sample_start_inds = np.ix_(sample_seeds)[0][those_satisfying_S1]

    # Find the indices of points that...
    local_seeds = np.logical_and.reduce((
        H['local_pt'],  # are localopt points
        H['dist_to_better_l'] > r_k,  # no better local point within r_k (L1)
        ~np.isinf(H['f']),  # have a non-infinity objective value
        ~np.isnan(H['f']),  # have a non-NaN objective value
        test_2_through_5,
        H['num_active_runs'] == 0,  # are not in an active run (L6)
        ~H['local_min']  # are not a local min (L7)
    ))

    local_start_inds2 = list(np.ix_(local_seeds)[0])

    # If paused is a field in H, don't start from paused points.
    if 'paused' in H.dtype.names:
        sample_start_inds = sample_start_inds[~H[sample_start_inds]['paused']]
        start_inds = list(sample_start_inds)+local_start_inds2
    else:
        start_inds = list(sample_start_inds)+local_start_inds2

    return start_inds


def look_in_history(x, Run_H, vector_return=False):
    """ See if Run['x_on_cube'][advance_local_run.pt_in_run] matches x,
    returning f or fvec, or saves x to advance_local_run.x_new if every point in Run_H has been
    checked.
    """

    if vector_return:
        to_return = 'fvec'
    else:
        if 'grad' in Run_H.dtype.names:
            to_return = ['f', 'grad']
        else:
            to_return = 'f'

    if advance_local_run.pt_in_run < len(Run_H):
        # Return the value in history to the localopt algorithm.
        assert np.allclose(x, Run_H['x_on_cube'][advance_local_run.pt_in_run], rtol=1e-08, atol=1e-08), \
            "History point does not match Localopt point"
        f_out = Run_H[to_return][advance_local_run.pt_in_run]
    else:
        if advance_local_run.pt_in_run == len(Run_H):
            # The history of points is exhausted. Save the requested point x to
            # x_new. x_new will be returned to the manager.
            advance_local_run.x_new[:] = x

        # Just in case the local opt method requests more points after a new
        # point has been identified.
        f_out = Run_H[to_return][-1]

    advance_local_run.pt_in_run += 1

    return f_out


def calc_rk(n, n_s, rk_const, lhs_divisions=0):
    """ Calculate the critical distance r_k """

    if lhs_divisions == 0:
        r_k = rk_const*(log(n_s)/n_s)**(1/n)
    else:
        k = np.floor(n_s/lhs_divisions).astype(int)
        if k <= 1:  # to prevent r_k=0
            r_k = np.inf
        else:
            r_k = rk_const*(log(k)/k)**(1/n)

    return r_k


def initialize_APOSMM(H, gen_specs):
    """
    Computes common values every time that APOSMM is reinvoked

    .. seealso::
        `start_persistent_local_opt_gens.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/alloc_funcs/start_persistent_local_opt_gens.py>`_
    """

    user_specs = gen_specs['user']
    n = len(user_specs['ub'])

    if 'single_component_at_a_time' in user_specs and user_specs['single_component_at_a_time']:
        c_flag = True
    else:
        c_flag = False

    if c_flag:
        # Get the pt_id for non-nan, returned points
        pt_ids = H['pt_id'][np.logical_and(H['returned'], ~np.isnan(H['f_i']))]
        _, counts = np.unique(pt_ids, return_counts=True)
        n_s = np.sum(counts == user_specs['components'])
    else:
        # Number of returned sampled points (excluding nans)
        n_s = np.sum(np.logical_and.reduce((~np.isnan(H['f']), ~H['local_pt'], H['returned'])))

    # Rather than build up a large output, we will just make changes in the
    # given H, and then send back the rows corresponding to updated H entries.
    Out = np.empty(0, dtype=gen_specs['out'])

    if 'rk_const' in user_specs:
        rk_c = user_specs['rk_const']
    else:
        rk_c = ((gamma(1+(n/2.0))*5.0)**(1.0/n))/sqrt(pi)

    if 'lhs_divisions' in user_specs:
        ld = user_specs['lhs_divisions']
    else:
        ld = 0

    if 'mu' in user_specs:
        mu = user_specs['mu']
    else:
        mu = 1e-4

    if 'nu' in user_specs:
        nu = user_specs['nu']
    else:
        nu = 0

    if n_s > 0:
        r_k = calc_rk(n, n_s, rk_c, ld)
    else:
        r_k = np.inf

    return n, n_s, c_flag, Out, r_k, mu, nu


def display_exception(e):
    print(e.__doc__)
    print(e.args)
    _, _, tb = sys.exc_info()
    traceback.print_tb(tb)  # Fixed format
    tb_info = traceback.extract_tb(tb)
    filename, line, func, text = tb_info[-1]
    print('An error occurred on line {} of function {} with statement {}'.format(line, func, text), flush=True)

    # PETSc/TAO errors are printed in the following manner:
    if hasattr(e, '_traceback_'):
        print('The error was:')
        for i in e._traceback_:
            print(i, flush=True)


# if __name__ == "__main__":
#     [H,gen_specs,persis_info] = [np.load('H20.npz')[i] for i in ['H','gen_specs','persis_info']]
#     gen_specs = gen_specs.item()
#     persis_info = persis_info.item()
#     import ipdb; ipdb.set_trace()
#     aposmm_logic(H,persis_info,gen_specs,{})



"""
Runs libEnsemble run-lines for adaptive workers with persistent gen.

Default setup is designed to run on 4*N + 1 workers - to modify, change total_nodes.
where one worker is a zero-resource persistent gen.

Execute via one of the following commands (e.g. 9 workers):
   mpiexec -np 10 python3 test_runlines_adaptive_workers_persistent.py

This is a dry run test, mocking up the nodes available. To test the run-lines
requires running a fixed, rather than random number of resource sets for a given sim_id.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 10

import numpy as np
import pkg_resources

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_with_variable_resources as sim_f

from libensemble.gen_funcs.persistent_uniform_sampling import uniform_random_sample_with_variable_resources as gen_f

# from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f

from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor
from libensemble.tests.regression_tests.common import create_node_file

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['zero_resource_workers'] = [1]
num_gens = len(libE_specs['zero_resource_workers'])
total_nodes = (nworkers - num_gens) // 4  # 4 resourced workers per node.

if total_nodes == 1:
    max_rsets = 4  # Up to one node
else:
    max_rsets = 6  # Will expand to 2 full nodes

sim_app = pkg_resources.resource_filename('libensemble.sim_funcs', 'helloworld.py')
exctr = MPIExecutor()
exctr.register_app(full_path=sim_app, app_name='helloworld')

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
    'user': {'dry_run': True},
}

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['x', 'f', 'sim_id'],
    'out': [('priority', float), ('resource_sets', int), ('x', float, n), ('x_on_cube', float, n)],
    'user': {
        'initial_batch_size': nworkers - 1,
        'max_resource_sets': max_rsets,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {
    'alloc_f': alloc_f,
    'user': {'give_all_with_same_priority': False},
}

comms = libE_specs['comms']
node_file = 'nodelist_adaptive_workers_persistent_comms_' + str(comms) + '_wrks_' + str(nworkers)
if is_manager:
    create_node_file(num_nodes=total_nodes, name=node_file)

if comms == 'mpi':
    libE_specs['mpi_comm'].Barrier()

libE_specs['resource_info'] = {
    'cores_on_node': (16, 64),  # Tuple (physical cores, logical cores)
    'node_file': node_file,
}  # Name of file containing a node-list

persis_info = add_unique_random_streams({}, nworkers + 1)
exit_criteria = {'sim_max': 40, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(
    sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs, alloc_specs=alloc_specs
)

if is_manager:
    assert flag == 0
    save_libE_output(H, persis_info, __file__, nworkers)



"""
This file contains many common subroutines used in distributed optimization
libraries, including collecting all the sum of {f_i}'s, collecting the
gradients, and conducting the consensus step (i.e., take linear combination of
your neighbors' $x$ values.
"""

import nlopt
import numpy as np
import scipy.sparse as spp
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport


def print_final_score(x, f_i_idxs, gen_specs, libE_info):
    """ This function is called by a gen so that the alloc will collect
        all the {f_i}'s and print their sum.

    Parameters
    ----------
    - x : np.ndarray
        Input solution vector
    - f_i_idxs : np.ndarray
        Which {f_i}'s this calling gen is responsible for
    - gen_specs, libE_info :
        Used to communicate
    """
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # evaluate { f_i(x) } first
    H_o = np.zeros(len(f_i_idxs), dtype=gen_specs['out'])
    H_o['x'][:] = x
    H_o['consensus_pt'][:] = False
    H_o['obj_component'][:] = f_i_idxs
    H_o['get_grad'][:] = False
    H_o = np.reshape(H_o, newshape=(-1,))

    tag, Work, calc_in = ps.send_recv(H_o)

    if tag in [PERSIS_STOP, STOP_TAG]:
        return

    f_is = calc_in['f_i']
    F_i = np.sum(f_is)

    # get alloc to print sum of f_i
    H_o = np.zeros(1, dtype=gen_specs['out'])
    H_o['x'][0] = x
    H_o['f_i'][0] = F_i
    H_o['eval_pt'][0] = True
    H_o['consensus_pt'][0] = True

    ps.send_recv(H_o)


def get_func_or_grad(x, f_i_idxs, gen_specs, libE_info, get_grad):
    """ This function is called by a gen to retrieve the function or gradient
        of the sum of {f_i}'s via the sim.

    Parameters
    ----------
    - x : np.ndarray
        Input solution vector
    - f_i_idxs : np.ndarray
        Which {f_i}'s this calling gen is responsible for
    - gen_specs, libE_info :
        Used to communicate
    - get_grad : bool
        True if we want gradient, otherwise returns function eval
    """
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    H_o = np.zeros(len(f_i_idxs), dtype=gen_specs['out'])
    H_o['x'][:] = x
    H_o['consensus_pt'][:] = False
    H_o['obj_component'][:] = f_i_idxs
    H_o['get_grad'][:] = get_grad
    H_o = np.reshape(H_o, newshape=(-1,))      # unfold into 1d array

    tag, Work, calc_in = ps.send_recv(H_o)

    if tag in [STOP_TAG, PERSIS_STOP]:
        return tag, None

    if get_grad:
        gradf_is = calc_in['gradf_i']
        gradf = np.sum(gradf_is, axis=0)
        return tag, gradf
    else:
        f_is = calc_in['f_i']
        f = np.sum(f_is)
        return tag, f


def get_func(x, f_i_idxs, gen_specs, libE_info):
    return get_func_or_grad(x, f_i_idxs, gen_specs, libE_info, get_grad=False)


def get_grad(x, f_i_idxs, gen_specs, libE_info):
    return get_func_or_grad(x, f_i_idxs, gen_specs, libE_info, get_grad=True)


def get_grad_locally(x, f_i_idxs, df):
    """ This function is called by a gen to locally compute gradients of
        the sum of {f_i}'s. Unlike `get_grad`, this function does not
        use the sim, but instead evaluates the gradient using the input @df.

    Parameters
    ----------
    - x : np.ndarray
        Input solution vector
    - f_i_idxs : np.ndarray
        Which {f_i}'s this calling gen is responsible for
    - df : func
        Function that returns gradient. Must take in as parameters input @x and
        index @i (i.e., which f_i to take gradient of)
    """
    gradf = np.zeros(len(x), dtype=float)
    for i in f_i_idxs:
        gradf += df(x, i)

    return gradf


def get_neighbor_vals(x, local_gen_id, A_gen_ids_no_local, gen_specs, libE_info):
    """ Sends local gen data (@x) and retrieves neighbors local data.
        Sorts the data so the gen ids are in increasing order

    Parameters
    ----------
    x : np.ndarray
        - local input variable

    local_gen_id : int
        - this gen's gen_id

    A_gen_ids_local : int
        - expected neighbor's gen ids, not including local gen id

    gen_specs, libE_info : ?
        - objects to communicate and construct mini History array

    Returns
    -------
    X : np.ndarray
        - 2D array of neighbors and local x values sorted by gen_ids
    """
    H_o = np.zeros(1, dtype=gen_specs['out'])
    H_o['x'][0] = x
    H_o['consensus_pt'][0] = True
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    tag, Work, calc_in = ps.send_recv(H_o)
    if tag in [STOP_TAG, PERSIS_STOP]:
        return tag, None

    neighbor_X = calc_in['x']
    neighbor_gen_ids = calc_in['gen_worker']

    assert local_gen_id not in neighbor_gen_ids, 'Local data should not be ' + \
                                                 'sent back from manager'
    assert np.array_equal(A_gen_ids_no_local, neighbor_gen_ids), 'Expected ' + \
        'gen_ids {}, received {}'.format(A_gen_ids_no_local, neighbor_gen_ids)

    X = np.vstack((neighbor_X, x))
    gen_ids = np.append(neighbor_gen_ids, local_gen_id)

    # sort data (including local) in corresponding gen_id increasing order
    X[:] = X[np.argsort(gen_ids)]

    return tag, X


def get_consensus_gradient(x, gen_specs, libE_info):
    """ Sends local gen data (@x) and retrieves neighbors local data,
        and takes sum of the neighbors' x's, which is equivalent to taking
        the gradient of consensus term for this particular node/agent.

        This function is equivalent to the @get_neighbor_vals function, but is
        less general, i.e., when we need only take a sum rather than a linear
        combination of our neighbors.

    Parameters
    ----------
    x : np.ndarray
        - local input variable

    Returns
    -------
     : np.ndarray
        - Returns this node's corresponding gradient of consensus
    """
    tag = None
    H_o = np.zeros(1, dtype=gen_specs['out'])
    H_o['x'][0] = x
    H_o['consensus_pt'][0] = True
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    tag, Work, calc_in = ps.send_recv(H_o)

    if tag in [PERSIS_STOP, STOP_TAG]:
        return tag, np.zeros(len(x))

    neighbor_X = calc_in['x']
    num_neighbors = len(neighbor_X)
    grad_cons = (num_neighbors*x) - np.sum(neighbor_X, axis=0)

    return tag, grad_cons


def get_k_reach_chain_matrix(n, k):
    """ Constructs adjacency matrix for a chain matrix where the ith vertex can
        reach vertices that are at most @k distances from them (does not wrap around),
        where the distance is based on the absolute difference between vertices'
        indexes.
    """
    assert 1 <= k <= n-1

    half_of_diagonals = [np.ones(n-k+j) for j in range(k)]
    half_of_indices = np.arange(1, k+1)
    all_of_diagonals = half_of_diagonals + half_of_diagonals[::-1]
    all_of_indices = np.append(-half_of_indices[::-1], half_of_indices)
    A = spp.csr_matrix(spp.diags(all_of_diagonals, all_of_indices))
    return A


def get_doubly_stochastic(A):
    """ Generates a doubly stochastic matrix where
    (i) S_ii > 0 for all i
    (ii) S_ij > 0 if and only if (i, j) in E

    Parameters
    ----------
    A : np.ndarray
        - adjacency matrix

    Returns
    -------
    x : scipy.sparse.csr_matrix
    """
    np.random.seed(0)
    n = A.shape[0]
    x = np.multiply(A.toarray() != 0, np.random.random((n, n)))
    x = x + np.diag(np.random.random(n) + 1e-4)

    rsum = np.zeros(n)
    csum = np.zeros(n)
    tol = 1e-15

    while (np.any(np.abs(rsum - 1) > tol)) | (np.any(np.abs(csum - 1) > tol)):
        x = x / x.sum(0)
        x = x / x.sum(1)[:, np.newaxis]
        rsum = x.sum(1)
        csum = x.sum(0)

    X = spp.csr_matrix(x)
    return X


"""
The remaining functions below don't help with consensus, but help with the
regression tests. One can move these functions to a different Python file
if need be.
"""


def readin_csv(fname):
    """ Parses breast-cancer dataset
        (http://archive.ics.uci.edu/ml/datasets/breast+cancer+wisconsin+%28diagnostic%29)
        for SVM.

    Parameters
    ----------
    - fname : str
        file name containing data

    Returns
    -------
    - labels : np.ndarray, (m,)
        1D with the label of each vector
    - datas : np.ndarray (2D), (m, n)
        2D array (matrix) with the collection of dataset
    """
    n = 569
    try:
        fp = open(fname, 'r+')
    except FileNotFoundError:
        print("# Missing file 'wdbc.data' (must be placed in same directory where you run 'mpirun -np ...'). "
              "File can be downloaded from https://tinyurl.com/3a2ttj5a. "
              "Creating artificial dataset instead.")

        m = 30
        np.random.seed(0)
        label = 2*np.random.randint(low=0, high=1, size=m)-1
        datas = (5*np.random.random(size=(n, m))).astype('int')

        return label, datas

    label = np.zeros(n, dtype='int')
    datas = np.zeros((n, 30))
    i = 0

    for line in fp.readlines():
        line = line.rsplit()[0]
        data = line.split(',')
        label[i] = data[1] == 'M'
        datas[i, :] = [float(val) for val in data[2:32]]
        i += 1

    assert i == n, 'Expected {} datapoints, recorded {}'.format(n, i)

    return label, datas


def gm_opt(b, m):
    """ Computes optimal geometric median score

    Parameters
    ----------
    - b : np.ndarray, (m*n,)
        1D array concatenating @m vectors of size @n, i.e., [x_1, x_2,..., x_m]
    - m : int
        number of vectors
    """

    n = len(b)//m
    assert len(b) == m*n

    ones_m = np.ones((m, 1))

    def obj_fn(x, B, m):
        X = np.outer(ones_m, x.T)
        return (1/m) * np.sum(np.linalg.norm(X-B, 2, axis=1))

    B = np.reshape(b, newshape=(m, n))

    opt = nlopt.opt(nlopt.LN_COBYLA, n)
    opt.set_min_objective(lambda beta, grad: obj_fn(beta, B, m))
    opt.set_xtol_rel(1e-8)

    opt.optimize(np.zeros(n))
    minf = opt.last_optimum_value()

    return minf


def regls_opt(X, y, c, reg=None):
    """ Computes optimal linear regression with l2 regularization

    Parameters
    ----------
    - X, y : np.ndarray
        2D matrix, 1D matrix, where we want to solve optimally for theta so that
        $y \approx X.dot(theta)$
    - c : float
        Scalar term for regularization
    - reg : str
        Denotes which regularization to use. Either 'l1', 'l2', or None
    """
    if reg == 'l1':
        p = 1
    elif reg == 'l2':
        p = 2
    elif reg is None:
        p = -1
    else:
        assert False, 'illegal regularization "{}"'.format(reg)

    def obj_fn(X, y, beta, c, p):
        m = X.shape[0]
        if p == 1:
            return (1/m) * np.linalg.norm(X @ beta - y, ord=2)**2 + c * np.linalg.norm(beta, ord=1)
        return (1/m) * np.linalg.norm(X @ beta - y, ord=2)**2 + c * np.linalg.norm(beta, ord=2)**2

    # already X.T
    d, m = X.shape

    opt = nlopt.opt(nlopt.LN_COBYLA, d)
    opt.set_min_objective(lambda beta, grad: obj_fn(X.T, y, beta, c, p))
    opt.set_xtol_rel(1e-8)

    opt.optimize(np.zeros(d))
    minf = opt.last_optimum_value()

    return minf


def log_opt(X, y, c, reg=None):
    """ Computes optimal linear regression with l2 regularization. See, for
        reference,
        https://www.cvxpy.org/examples/machine_learning/logistic_regression.html

    Parameters
    ----------
    - X, y : np.ndarray
        2D matrix, 1D matrix, defining the logisitic regression problem
    - c : float
        Scalar term for regularization
    - reg : str
        Denotes which regularization to use. Either 'l1', 'l2', or None
    """
    assert reg == 'l2', "Only l2 regularization allowed"
    # if reg == 'l1':
    #     p = 1
    # elif reg == 'l2':
    p = 2
    # elif reg is None:
    #     p = 0
    # else:
    #     assert False, 'illegal regularization mode, "{}"'.format(reg)

    def obj_fn(X, y, beta, c, p):
        m = X.shape[0]
        # if p == 0:
        #     reg = 0
        # if p == 1:
        #     reg = c * np.linalg.norm(beta, 1)
        # elif p == 2:
        reg = c * np.linalg.norm(beta, 2)**2
        # Note that, cp.logistic(x) == log(1+e^x)
        return (1/m) * np.sum(np.log(1+np.exp(np.multiply(-y, X @ beta)))) + reg

    d, m = X.shape

    opt = nlopt.opt(nlopt.LN_COBYLA, d)
    opt.set_min_objective(lambda beta, grad: obj_fn(X.T, y, beta, c, p))
    opt.set_xtol_rel(1e-8)

    opt.optimize(np.zeros(d))
    minf = opt.last_optimum_value()

    return minf


def svm_opt(X, b, c, reg='l1'):
    """ Computes optimal support vector machine (SVM) with l1 regularization.

    Parameters
    ----------
    - X, b : np.ndarray
        2D matrix, 1D matrix, defining the SVM problem
    - c : float
        Scalar term for regularization
    - reg : str
        Denotes which regularization to use. Either 'l1', 'l2', or None
    """
    assert reg == 'l1', "Only l1 regularization allowed"

    # if reg == 'l1':
    p = 1
    # elif reg == 'l2':
    #     p = 2
    # elif reg is None:
    #     p = 0
    # else:
    #     assert False, 'illegal regularization mode, "{}"'.format(reg)

    def obj_fn(X, b, theta, c, p):
        # if p == 0:
        #     reg = 0
        # if p == 1:
        reg = c * np.linalg.norm(theta, 1)
        # if p == 2:
        #     reg = c * np.linalg.norm(theta, 2)**2
        return np.sum(np.maximum(0, 1-np.multiply(b, X @ theta))) + reg

    d, m = X.shape
    opt = nlopt.opt(nlopt.LN_COBYLA, d)
    opt.set_min_objective(lambda theta, grad: obj_fn(X.T, b, theta, c, p))
    opt.set_xtol_rel(1e-8)

    opt.optimize(np.zeros(d))
    minf = opt.last_optimum_value()

    return minf



"""
Runs libEnsemble testing the MPI Runners command creation with multiple and uneven nodes per worker.

This test must be run on a number of workers >= 2.

Execute via one of the following commands (e.g. 5 workers):
   mpiexec -np 6 python3 test_mpi_runners_supernode_uneven.py
   python3 test_mpi_runners_supernode_uneven.py --nworkers 5 --comms local
"""

import numpy as np

from libensemble.libE import libE
from libensemble.sim_funcs.run_line_check import runline_check_by_worker as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor
from libensemble.tests.regression_tests.common import create_node_file
from libensemble import logger

# logger.set_level('DEBUG')  # For testing the test
logger.set_level('INFO')

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 3 4 5

nworkers, is_manager, libE_specs, _ = parse_args()
rounds = 1
sim_app = '/path/to/fakeapp.x'
comms = libE_specs['comms']

libE_specs['dedicated_mode'] = True
libE_specs['enforce_worker_core_bounds'] = True

# To allow visual checking - log file not used in test
log_file = 'ensemble_mpi_runners_supernode_uneven_comms_' + str(comms) + '_wrks_' + str(nworkers) + '.log'
logger.set_filename(log_file)

nodes_per_worker = 2.5

# For varying size test - relate node count to nworkers
nsim_workers = nworkers
comms = libE_specs['comms']
node_file = 'nodelist_mpi_runners_supernode_uneven_comms_' + str(comms) + '_wrks_' + str(nworkers)
nnodes = int(nsim_workers * nodes_per_worker)

# Mock up system
custom_resources = {
    'cores_on_node': (16, 64),  # Tuple (physical cores, logical cores)
    'node_file': node_file,
}  # Name of file containing a node-list
libE_specs['resource_info'] = custom_resources

if is_manager:
    create_node_file(num_nodes=nnodes, name=node_file)

if comms == 'mpi':
    libE_specs['mpi_comm'].Barrier()

# Create executor and register sim to it.
exctr = MPIExecutor(custom_info={'mpi_runner': 'srun'})
exctr.register_app(full_path=sim_app, calc_type='sim')

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': [],
    'out': [('x', float, (n,))],
    'user': {
        'gen_batch_size': 20,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)
exit_criteria = {'sim_max': (nsim_workers) * rounds}

# Each worker has either 3 or 2 nodes. Basic test list for portable options
test_list_base = [
    {'testid': 'base1'},  # Give no config and no extra_args
]

# Example: On 2 workers, runlines should be ...
# (one workers has 3 nodes, the other 2 - does not split 2.5 nodes each).
# [w1]: srun -w node-1,node-2,node-3 --ntasks 48 --nodes 3 --ntasks-per-node 16 /path/to/fakeapp.x --testid base1
# [w2]: srun -w node-4,node-5 --ntasks 32 --nodes 2 --ntasks-per-node 16 /path/to/fakeapp.x --testid base1

srun_p1 = 'srun -w '
srun_p2 = ' --ntasks '
srun_p3 = ' --nodes '
srun_p4 = ' --ntasks-per-node 16 /path/to/fakeapp.x --testid base1'

exp_tasks = []
exp_srun = []

# Hard coding an example for 2 nodes to avoid replicating general logic in libEnsemble.
low_npw = nnodes // nsim_workers
high_npw = nnodes // nsim_workers + 1

nodelist = []
for i in range(1, nnodes + 1):
    nodelist.append('node-' + str(i))

inode = 0
for i in range(nsim_workers):
    if i < (nsim_workers // 2):
        npw = high_npw
    else:
        npw = low_npw
    nodename = ','.join(nodelist[inode : inode + npw])
    inode += npw
    ntasks = 16 * npw
    loc_nodes = npw
    exp_tasks.append(ntasks)
    exp_srun.append(srun_p1 + str(nodename) + srun_p2 + str(ntasks) + srun_p3 + str(loc_nodes) + srun_p4)

test_list = test_list_base
exp_list = exp_srun
sim_specs['user'] = {
    'tests': test_list,
    'expect': exp_list,
}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

# All asserts are in sim func



"""
libEnsemble manager routines
============================
"""

import sys
import os
import glob
import logging
import socket
import traceback
import numpy as np

from libensemble.utils.timer import Timer
from libensemble.message_numbers import \
    EVAL_SIM_TAG, FINISHED_PERSISTENT_SIM_TAG, \
    EVAL_GEN_TAG, FINISHED_PERSISTENT_GEN_TAG, \
    STOP_TAG, UNSET_TAG, PERSIS_STOP, \
    WORKER_KILL, WORKER_KILL_ON_ERR, WORKER_KILL_ON_TIMEOUT, \
    TASK_FAILED, WORKER_DONE, \
    MAN_SIGNAL_FINISH, MAN_SIGNAL_KILL
from libensemble.message_numbers import calc_type_strings
from libensemble.comms.comms import CommFinishedException
from libensemble.worker import WorkerErrMsg
from libensemble.output_directory import EnsembleDirectory
from libensemble.tools.tools import _USER_CALC_DIR_WARNING
from libensemble.resources.resources import Resources
from libensemble.tools.tools import _PERSIS_RETURN_WARNING
from libensemble.tools.fields_keys import protected_libE_fields
import cProfile
import pstats
import copy

if tuple(np.__version__.split('.')) >= ('1', '15'):
    from numpy.lib.recfunctions import repack_fields

logger = logging.getLogger(__name__)
# For debug messages - uncomment
# logger.setLevel(logging.DEBUG)


class ManagerException(Exception):
    """Exception raised by the Manager"""


class WorkerException(Exception):
    """Exception raised on abort signal from worker"""


class LoggedException(Exception):
    """Raise exception for handling without re-logging"""


def report_worker_exc(wrk_exc=None):
    """Write worker exception to log"""
    if wrk_exc is not None:
        from_line, msg, exc = wrk_exc.args
        logger.error("---- {} ----".format(from_line))
        logger.error("Message: {}".format(msg))
        logger.error(exc)


def manager_main(hist, libE_specs, alloc_specs,
                 sim_specs, gen_specs, exit_criteria, persis_info, wcomms=[]):
    """Manager routine to coordinate the generation and simulation evaluations

    Parameters
    ----------

    hist: :obj:`History`
        A libEnsemble history type object.

    libE_specs: :obj:`dict`
        Specifications for libEnsemble

    alloc_specs: :obj:`dict`
        Specifications for the allocation function

    sim_specs: :obj:`dict`
        Specifications for the simulation function

    gen_specs: :obj:`dict`
        Specifications for the generator function

    exit_criteria: :obj:`dict`
        Criteria for libEnsemble to stop a run

    persis_info: :obj:`dict`
        Persistent information to be passed between user functions

    wcomms: :obj:`list`, optional
        A list of comm type objects for each worker. Default is an empty list.
    """
    if libE_specs.get('profile'):
        pr = cProfile.Profile()
        pr.enable()

    if 'in' not in gen_specs:
        gen_specs['in'] = []

    # Send dtypes to workers
    if 'repack_fields' in globals():
        dtypes = {EVAL_SIM_TAG: repack_fields(hist.H[sim_specs['in']]).dtype,
                  EVAL_GEN_TAG: repack_fields(hist.H[gen_specs['in']]).dtype}
    else:
        dtypes = {EVAL_SIM_TAG: hist.H[sim_specs['in']].dtype,
                  EVAL_GEN_TAG: hist.H[gen_specs['in']].dtype}

    for wcomm in wcomms:
        wcomm.send(0, dtypes)

    # Set up and run manager
    mgr = Manager(hist, libE_specs, alloc_specs,
                  sim_specs, gen_specs, exit_criteria, wcomms)
    result = mgr.run(persis_info)

    if libE_specs.get('profile'):
        pr.disable()
        profile_stats_fname = 'manager.prof'

        with open(profile_stats_fname, 'w') as f:
            ps = pstats.Stats(pr, stream=f).sort_stats('cumulative')
            ps.print_stats()

    return result


def filter_nans(array):
    """Filters out NaNs from a numpy array"""
    return array[~np.isnan(array)]


_WALLCLOCK_MSG_ALL_RETURNED = """
Termination due to elapsed_wallclock_time has occurred.
All completed work has been returned.
Posting kill messages for all workers.
"""

_WALLCLOCK_MSG_ACTIVE = """
Termination due to elapsed_wallclock_time has occurred.
Some issued work has not been returned.
Posting kill messages for all workers.
"""


class Manager:
    """Manager class for libensemble."""

    worker_dtype = [('worker_id', int),
                    ('active', int),
                    ('persis_state', int),
                    ('active_recv', int),
                    ('zero_resource_worker', bool)]

    def __init__(self, hist, libE_specs, alloc_specs,
                 sim_specs, gen_specs, exit_criteria,
                 wcomms=[]):
        """Initializes the manager"""
        timer = Timer()
        timer.start()
        self.date_start = timer.date_start.replace(' ', '_')
        self.safe_mode = libE_specs.get('safe_mode', True)
        self.kill_canceled_sims = libE_specs.get('kill_canceled_sims', True)
        self.hist = hist
        self.libE_specs = libE_specs
        self.alloc_specs = alloc_specs
        self.sim_specs = sim_specs
        self.gen_specs = gen_specs
        self.exit_criteria = exit_criteria
        self.elapsed = lambda: timer.elapsed
        self.wcomms = wcomms
        self.WorkerExc = False
        self.persis_pending = []
        self.W = np.zeros(len(self.wcomms), dtype=Manager.worker_dtype)
        self.W['worker_id'] = np.arange(len(self.wcomms)) + 1
        self.term_tests = \
            [(2, 'elapsed_wallclock_time', self.term_test_wallclock),
             (1, 'sim_max', self.term_test_sim_max),
             (1, 'gen_max', self.term_test_gen_max),
             (1, 'stop_val', self.term_test_stop_val)]

        temp_EnsembleDirectory = EnsembleDirectory(libE_specs=libE_specs)
        self.resources = Resources.resources
        if self.resources is not None:
            for wrk in self.W:
                if wrk['worker_id'] in self.resources.glob_resources.zero_resource_workers:
                    wrk['zero_resource_worker'] = True

        try:
            temp_EnsembleDirectory.make_copyback_check()
        except OSError as e:  # Ensemble dir exists and isn't empty.
            logger.manager_warning(_USER_CALC_DIR_WARNING.format(temp_EnsembleDirectory.prefix))
            self._kill_workers()
            raise ManagerException("Manager errored on initialization",
                                   "Ensemble directory already existed and wasn't empty.", e)

    # --- Termination logic routines

    def term_test_wallclock(self, max_elapsed):
        """Checks against wallclock timeout"""
        return self.elapsed() >= max_elapsed

    def term_test_sim_max(self, sim_max):
        """Checks against max simulations"""
        return self.hist.returned_count >= sim_max + self.hist.returned_offset

    def term_test_gen_max(self, gen_max):
        """Checks against max generator calls"""
        return self.hist.index >= gen_max + self.hist.given_back_offset

    def term_test_stop_val(self, stop_val):
        """Checks against stop value criterion"""
        key, val = stop_val
        H = self.hist.H
        return np.any(filter_nans(H[key][H['returned']]) <= val)

    def term_test(self, logged=True):
        """Checks termination criteria"""
        for retval, key, testf in self.term_tests:
            if key in self.exit_criteria:
                if testf(self.exit_criteria[key]):
                    if logged:
                        logger.info("Term test tripped: {}".format(key))
                    return retval
        return 0

    # --- Low-level communication routines

    def _kill_workers(self):
        """Kills the workers"""
        for w in self.W['worker_id']:
            self.wcomms[w-1].send(STOP_TAG, MAN_SIGNAL_FINISH)

    # --- Checkpointing logic

    def _save_every_k(self, fname, count, k):
        """Saves history every kth step"""
        count = k*(count//k)
        filename = fname.format(self.date_start, count)
        if not os.path.isfile(filename) and count > 0:
            for old_file in glob.glob(fname.format(self.date_start, '*')):
                os.remove(old_file)
            np.save(filename, self.hist.H)

    def _save_every_k_sims(self):
        """Saves history every kth sim step"""
        self._save_every_k('libE_history_for_run_starting_{}_after_sim_{}.npy',
                           self.hist.returned_count,
                           self.libE_specs['save_every_k_sims'])

    def _save_every_k_gens(self):
        """Saves history every kth gen step"""
        self._save_every_k('libE_history_for_run_starting_{}_after_gen_{}.npy',
                           self.hist.index,
                           self.libE_specs['save_every_k_gens'])

    # --- Handle outgoing messages to workers (work orders from alloc)

    def _check_work_order(self, Work, w):
        """Checks validity of an allocation function order
        """
        assert w != 0, "Can't send to worker 0; this is the manager."
        if self.W[w-1]['active_recv']:
            assert 'active_recv' in Work['libE_info'], \
                "Messages to a worker in active_recv mode should have active_recv"\
                "set to True in libE_info. Work['libE_info'] is {}".format(Work['libE_info'])
        else:
            assert self.W[w-1]['active'] == 0, \
                "Allocation function requested work be sent to worker %d, an "\
                "already active worker." % w
        work_rows = Work['libE_info']['H_rows']
        if len(work_rows):
            work_fields = set(Work['H_fields'])
            assert len(work_fields), \
                "Allocation function requested rows={} be sent to worker={}, "\
                "but requested no fields to be sent.".format(work_rows, w)
            hist_fields = self.hist.H.dtype.names
            diff_fields = list(work_fields.difference(hist_fields))
            assert not diff_fields, \
                "Allocation function requested invalid fields {}" \
                "be sent to worker={}.".format(diff_fields, w)

    def _set_resources(self, Work, w):
        """Check rsets given in Work match rsets assigned in resources.

        If rsets are not assigned, then assign using default mapping
        """
        resource_manager = self.resources.resource_manager
        rset_req = Work['libE_info'].get('rset_team')

        if rset_req is None:
            rset_team = []
            default_rset = resource_manager.index_list[w-1]
            if default_rset is not None:
                rset_team.append(default_rset)
            Work['libE_info']['rset_team'] = rset_team

        resource_manager.assign_rsets(Work['libE_info']['rset_team'], w)

    def _freeup_resources(self, w):
        """Free up resources assigned to the worker"""

        if self.resources:
            self.resources.resource_manager.free_rsets(w)

    def _send_work_order(self, Work, w):
        """Sends an allocation function order to a worker
        """

        logger.debug("Manager sending work unit to worker {}".format(w))

        if self.resources:
            self._set_resources(Work, w)

        self.wcomms[w-1].send(Work['tag'], Work)

        work_rows = Work['libE_info']['H_rows']
        work_name = calc_type_strings[Work['tag']]
        logger.debug("Manager sending {} work to worker {}. Rows {}".
                     format(work_name, w, EnsembleDirectory.extract_H_ranges(Work) or None))
        if len(work_rows):
            if 'repack_fields' in globals():
                new_dtype = [(name, self.hist.H.dtype.fields[name][0]) for name in Work['H_fields']]
                H_to_be_sent = np.empty(len(work_rows), dtype=new_dtype)
                for i, row in enumerate(work_rows):
                    H_to_be_sent[i] = repack_fields(self.hist.H[Work['H_fields']][row])
                # H_to_be_sent = repack_fields(self.hist.H[Work['H_fields']])[work_rows]
                self.wcomms[w-1].send(0, H_to_be_sent)
            else:
                self.wcomms[w-1].send(0, self.hist.H[Work['H_fields']][work_rows])

    def _update_state_on_alloc(self, Work, w):
        """Updates a workers' active/idle status following an allocation order"""

        self.W[w-1]['active'] = Work['tag']
        if 'libE_info' in Work:
            if 'persistent' in Work['libE_info']:
                self.W[w-1]['persis_state'] = Work['tag']
                if Work['libE_info'].get('active_recv', False):
                    self.W[w-1]['active_recv'] = Work['tag']
            else:
                assert 'active_recv' not in Work['libE_info'], \
                    "active_recv worker must also be persistent"

        work_rows = Work['libE_info']['H_rows']
        if Work['tag'] == EVAL_SIM_TAG:
            self.hist.update_history_x_out(work_rows, w)
        elif Work['tag'] == EVAL_GEN_TAG:
            self.hist.update_history_to_gen(work_rows)

    # --- Handle incoming messages from workers

    @staticmethod
    def _check_received_calc(D_recv):
        """Checks the type and status fields on a receive calculation"""
        calc_type = D_recv['calc_type']
        calc_status = D_recv['calc_status']
        assert calc_type in [EVAL_SIM_TAG, EVAL_GEN_TAG], \
            "Aborting, Unknown calculation type received. " \
            "Received type: {}".format(calc_type)
        assert calc_status in [FINISHED_PERSISTENT_SIM_TAG,
                               FINISHED_PERSISTENT_GEN_TAG,
                               UNSET_TAG,
                               PERSIS_STOP,
                               MAN_SIGNAL_FINISH,
                               MAN_SIGNAL_KILL,
                               WORKER_KILL_ON_ERR,
                               WORKER_KILL_ON_TIMEOUT,
                               WORKER_KILL,
                               TASK_FAILED,
                               WORKER_DONE], \
            "Aborting: Unknown calculation status received. " \
            "Received status: {}".format(calc_status)

    def _receive_from_workers(self, persis_info):
        """Receives calculation output from workers. Loops over all
        active workers and probes to see if worker is ready to
        communticate. If any output is received, all other workers are
        looped back over.
        """
        new_stuff = True
        while new_stuff and any(self.W['active']):
            new_stuff = False
            for w in self.W['worker_id'][self.W['active'] > 0]:
                if self.wcomms[w-1].mail_flag():
                    new_stuff = True
                    self._handle_msg_from_worker(persis_info, w)

        if 'save_every_k_sims' in self.libE_specs:
            self._save_every_k_sims()
        if 'save_every_k_gens' in self.libE_specs:
            self._save_every_k_gens()
        return persis_info

    def _update_state_on_worker_msg(self, persis_info, D_recv, w):
        """Updates history and worker info on worker message
        """
        calc_type = D_recv['calc_type']
        calc_status = D_recv['calc_status']
        Manager._check_received_calc(D_recv)

        if w not in self.persis_pending and not self.W[w-1]['active_recv']:
            self.W[w-1]['active'] = 0

        if calc_status in [FINISHED_PERSISTENT_SIM_TAG,
                           FINISHED_PERSISTENT_GEN_TAG]:
            final_data = D_recv.get('calc_out', None)
            if isinstance(final_data, np.ndarray):
                if calc_status is FINISHED_PERSISTENT_GEN_TAG and self.libE_specs.get('use_persis_return_gen', False):
                    self.hist.update_history_x_in(w, final_data, self.safe_mode)
                elif calc_status is FINISHED_PERSISTENT_SIM_TAG and self.libE_specs.get('use_persis_return_sim', False):
                    self.hist.update_history_f(D_recv, self.safe_mode)
                else:
                    logger.info(_PERSIS_RETURN_WARNING)
            self.W[w-1]['persis_state'] = 0
            if self.W[w-1]['active_recv']:
                self.W[w-1]['active'] = 0
                self.W[w-1]['active_recv'] = 0
            if w in self.persis_pending:
                self.persis_pending.remove(w)
                self.W[w-1]['active'] = 0
            self._freeup_resources(w)
        else:
            if calc_type == EVAL_SIM_TAG:
                self.hist.update_history_f(D_recv, self.safe_mode)
            if calc_type == EVAL_GEN_TAG:
                self.hist.update_history_x_in(w, D_recv['calc_out'], self.safe_mode)
                assert len(D_recv['calc_out']) or np.any(self.W['active']) or self.W[w-1]['persis_state'], \
                    "Gen must return work when is is the only thing active and not persistent."
            if 'libE_info' in D_recv and 'persistent' in D_recv['libE_info']:
                # Now a waiting, persistent worker
                self.W[w-1]['persis_state'] = calc_type
            else:
                self._freeup_resources(w)

        if 'persis_info' in D_recv and len(D_recv['persis_info']):
            persis_info[w].update(D_recv['persis_info'])

    def _handle_msg_from_worker(self, persis_info, w):
        """Handles a message from worker w
        """
        try:
            msg = self.wcomms[w-1].recv()
            tag, D_recv = msg
        except CommFinishedException:
            logger.debug("Finalizing message from Worker {}".format(w))
            return
        if isinstance(D_recv, WorkerErrMsg):
            self.W[w-1]['active'] = 0
            logger.debug("Manager received exception from worker {}".format(w))
            if not self.WorkerExc:
                self.WorkerExc = True
                self._kill_workers()
                raise WorkerException('Received error message from worker {}'.format(w),
                                      D_recv.msg, D_recv.exc)
        elif isinstance(D_recv, logging.LogRecord):
            logger.debug("Manager received a log message from worker {}".format(w))
            logging.getLogger(D_recv.name).handle(D_recv)
        else:
            logger.debug("Manager received data message from worker {}".format(w))
            self._update_state_on_worker_msg(persis_info, D_recv, w)

    def _kill_cancelled_sims(self):
        """Send kill signals to any sims marked as cancel_requested"""
        if self.kill_canceled_sims:
            kill_sim = self.hist.H['given'] & self.hist.H['cancel_requested'] \
                & ~self.hist.H['returned'] & ~self.hist.H['kill_sent']

            # Note that a return is still expected when running sims are killed
            if np.any(kill_sim):
                logger.debug('Manager sending kill signals to H indices {}'.format(np.where(kill_sim)))
                kill_ids = self.hist.H['sim_id'][kill_sim]
                kill_on_workers = self.hist.H['sim_worker'][kill_sim]
                for w in kill_on_workers:
                    self.wcomms[w-1].send(STOP_TAG, MAN_SIGNAL_KILL)
                    self.hist.H['kill_sent'][kill_ids] = True

    # --- Handle termination

    def _final_receive_and_kill(self, persis_info):
        """
        Tries to receive from any active workers.

        If time expires before all active workers have been received from, a
        nonblocking receive is posted (though the manager will not receive this
        data) and a kill signal is sent.
        """

        # Send a handshake signal to each persistent worker.
        if any(self.W['persis_state']):
            for w in self.W['worker_id'][self.W['persis_state'] > 0]:
                logger.debug("Manager sending PERSIS_STOP to worker {}".format(w))
                if 'final_fields' in self.libE_specs:
                    rows_to_send = self.hist.trim_H()['returned']
                    fields_to_send = self.libE_specs['final_fields']
                    H_to_send = self.hist.trim_H()[rows_to_send][fields_to_send]
                    self.wcomms[w-1].send(PERSIS_STOP, H_to_send)
                else:
                    self.wcomms[w-1].send(PERSIS_STOP, MAN_SIGNAL_KILL)
                if not self.W[w-1]['active']:
                    # Re-activate if necessary
                    self.W[w-1]['active'] = self.W[w-1]['persis_state']
                self.persis_pending.append(w)

        exit_flag = 0
        while (any(self.W['active']) or any(self.W['persis_state'])) and exit_flag == 0:
            persis_info = self._receive_from_workers(persis_info)
            if self.term_test(logged=False) == 2:
                # Elapsed Wallclock has expired
                if not any(self.W['persis_state']):
                    if any(self.W['active']):
                        logger.manager_warning(_WALLCLOCK_MSG_ACTIVE)
                    else:
                        logger.manager_warning(_WALLCLOCK_MSG_ALL_RETURNED)
                    exit_flag = 2
            if self.WorkerExc:
                exit_flag = 1

        self._kill_workers()
        return persis_info, exit_flag, self.elapsed()

    # --- Main loop

    def _sim_max_given(self):
        if 'sim_max' in self.exit_criteria:
            return self.hist.given_count >= self.exit_criteria['sim_max'] + self.hist.given_offset
        else:
            return False

    def _get_alloc_libE_info(self):
        """Selected statistics useful for alloc_f"""

        return {'any_idle_workers': any(self.W['active'] == 0),
                'exit_criteria': self.exit_criteria,
                'elapsed_time': self.elapsed(),
                'manager_kill_canceled_sims': self.kill_canceled_sims,
                'given_count': self.hist.given_count,
                'returned_count': self.hist.returned_count,
                'given_back_count': self.hist.given_back_count,
                'sim_max_given': self._sim_max_given(),
                'use_resource_sets': 'num_resource_sets' in self.libE_specs}

    def _alloc_work(self, H, persis_info):
        """
        Calls work allocation function from alloc_specs. Copies protected libE
        fields before the alloc_f call and ensures they weren't modified
        """
        if self.safe_mode:
            if 'repack_fields' in globals():
                saveH = repack_fields(H[protected_libE_fields], recurse=True)
            else:
                saveH = copy.deepcopy(H[protected_libE_fields])

        alloc_f = self.alloc_specs['alloc_f']
        output = alloc_f(self.W, H, self.sim_specs, self.gen_specs, self.alloc_specs,
                         persis_info, self._get_alloc_libE_info())

        if self.safe_mode:
            assert np.array_equal(saveH, H[protected_libE_fields]), "The allocation function modified protected fields"

        if len(output) == 2:
            output = output + ((0,))

        assert len(output) == 3, "alloc_f must return three outputs."
        assert isinstance(output[0], dict), "First alloc_f output must be a dictionary"
        assert isinstance(output[1], dict), "Second alloc_f output must be a dictionary"
        assert output[2] in [0, 1], "Third alloc_f output must be 0 or 1."

        return output

    def run(self, persis_info):
        """Runs the manager"""
        logger.info("Manager initiated on node {}".format(socket.gethostname()))
        logger.info("Manager exit_criteria: {}".format(self.exit_criteria))

        # Continue receiving and giving until termination test is satisfied
        try:
            while not self.term_test():
                self._kill_cancelled_sims()
                persis_info = self._receive_from_workers(persis_info)
                Work, persis_info, flag = self._alloc_work(self.hist.trim_H(),
                                                           persis_info)
                if flag:
                    break

                for w in Work:
                    if self._sim_max_given():
                        break
                    self._check_work_order(Work[w], w)
                    self._send_work_order(Work[w], w)
                    self._update_state_on_alloc(Work[w], w)
                assert self.term_test() or any(self.W['active'] != 0), \
                    "alloc_f did not return any work, although all workers are idle."
        except WorkerException as e:
            report_worker_exc(e)
            raise LoggedException(e.args[0], e.args[1]) from None
        except Exception as e:
            logger.error(traceback.format_exc())
            raise LoggedException(e.args) from None
        finally:
            # Return persis_info, exit_flag, elapsed time
            result = self._final_receive_and_kill(persis_info)
            sys.stdout.flush()
            sys.stderr.flush()
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# liEensemble documentation build configuration file, created by
# sphinx-quickstart on Fri Aug 18 11:52:31 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import datetime

exec(open('../libensemble/version.py').read())


if sys.version_info >= (3, 3):
    from unittest.mock import MagicMock
else:
    from mock import MagicMock

class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
            return MagicMock()

MOCK_MODULES = [
        'argparse',
        'deap',
        'dfols',
        'math',
        'mpi4py',
        'mpmath',
        'nlopt',
        'numpy',
        'numpy.lib',
        'numpy.lib.recfunctions',
        'numpy.linalg',
        'PETSc',
        'petsc4py',
        'psutil',
        'scipy',
        'scipy.io',
        'scipy.sparse',
        'scipy.spatial',
        'scipy.spatial.distance',
        'scipy.stats',
        'surmise.calibration',
        'surmise.emulation',
        'Tasmanian',
        'libensemble.gen_funcs.vtmop_libe'
        ]

sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

#from libensemble import *
#from libensemble.alloc_funcs import *
#from libensemble.gen_funcs import *
#from libensemble.sim_funcs import *


#sys.path.insert(0, os.path.abspath('.'))
sys.path.append(os.path.abspath('../libensemble'))
##sys.path.append(os.path.abspath('../libensemble'))
sys.path.append(os.path.abspath('../libensemble/alloc_funcs'))
sys.path.append(os.path.abspath('../libensemble/gen_funcs'))
sys.path.append(os.path.abspath('../libensemble/sim_funcs'))
sys.path.append(os.path.abspath('../libensemble/comms'))
sys.path.append(os.path.abspath('../libensemble/utils'))
sys.path.append(os.path.abspath('../libensemble/tools'))
sys.path.append(os.path.abspath('../libensemble/executors'))
sys.path.append(os.path.abspath('../libensemble/resources'))
# print(sys.path)

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = '2.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinxcontrib.bibtex',
              'sphinx.ext.autodoc',
              'sphinx.ext.napoleon',
              # 'sphinx.ext.autosectionlabel',
              # 'sphinx.ext.intersphinx',
              'sphinx.ext.imgconverter',
              'sphinx.ext.mathjax']
bibtex_bibfiles = ['references.bib']
bibtex_default_style = 'unsrt'
# autosectionlabel_prefix_document = True
# extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon', 'sphinx.ext.imgconverter']
#breathe_projects = { "libEnsemble": "../code/src/xml/" }
#breathe_default_project = "libEnsemble"

##breathe_projects_source = {"libEnsemble" : ( "../code/src/", ["libE.py", "test.cpp"] )}
#breathe_projects_source = {"libEnsemble" : ( "../code/src/", ["test.cpp","test2.cpp"] )}

autodoc_mock_imports = ["balsam"]
extlinks = {'duref': ('http://docutils.sourceforge.net/docs/ref/rst/'
                      'restructuredtext.html#%s', ''),
            'durole': ('http://docutils.sourceforge.net/docs/ref/rst/'
                       'roles.html#%s', ''),
            'dudir': ('http://docutils.sourceforge.net/docs/ref/rst/'
                      'directives.html#%s', '')}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Configure bibtex_bibfiles setting for sphinxcontrib-bibtex 2.0.0
bibtex_bibfiles = ['references.bib']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The latex toctree document.
latex_doc = 'latex_index'

# General information about the project.
project = 'libEnsemble'
copyright = str(datetime.now().year) + ' Argonne National Laboratory'
author = 'Jeffrey Larson, Stephen Hudson, Stefan M. Wild, David Bindel and John-Luke Navarro'
today_fmt = '%B %-d, %Y'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
#html_theme = 'default'
#html_theme = 'graphite'
# html_theme = 'sphinxdoc'
html_theme = 'sphinx_rtd_theme'

html_logo = './images/libE_logo_white.png'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {'navigation_depth': 3,
                      'collapse_navigation': False,
                      'logo_only': True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']
html_static_path = []

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
#html_sidebars = {
#    '**': [
#        'about.html',
#        'navigation.html',
#        'relations.html',  # needs 'show_related': True theme option to display
#        'searchbox.html',
#        'donate.html',
#    ]
#}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'libEnsembledoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # Sonny, Lenny, or Bjornstrup
    'fncychap': '\\usepackage[Lenny]{fncychap}',
    'extraclassoptions': 'openany',
    'preamble':
    r'''
    \protected\def\sphinxcrossref#1{\texttt{#1}}

    \newsavebox\mytempbox
    \definecolor{sphinxnoteBgColor}{RGB}{221,233,239}
    \renewenvironment{sphinxnote}[1]{%
    \begin{lrbox}{\mytempbox}%
    \begin{minipage}{\columnwidth}%
    \begin{sphinxlightbox}%
    \sphinxstrong{#1}}%
    {\end{sphinxlightbox}%
    \end{minipage}%
    \end{lrbox}%
    \colorbox{sphinxnoteBgColor}{\usebox{\mytempbox}}}
                ''',
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (latex_doc, 'libEnsemble.tex', 'libEnsemble User Manual',
     'Stephen Hudson, Jeffrey Larson, Stefan M. Wild, \\\\ \\hfill David Bindel, John-Luke Navarro', 'manual'),
]

latex_logo = 'images/libE_logo.png'

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (latex_doc, 'libEnsemble', 'libEnsemble User Manual',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (latex_doc, 'libEnsemble', 'libEnsemble User Manual',
     author, 'libEnsemble', 'One line description of project.',
     'Miscellaneous'),
]
